%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           ibmrs-registers.sl
% Title:          IBM RS 6000 PSL register allocation
% Author:         Winfried Neun , ZIB Berlin 
% Created:        9 November 1990
% Mode:           Lisp
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%	Define register names
%%
(loadtime (progn
(deflist '((r0 0)(r1 1)(r2 2)(r3 3)(r4 4)(r5 5)(r6 6)(r7 7)
	   (r8 8)(r9 9)(r10 10)(r11 11)(r12 12)(r13 13)(r14 14)
	   (r15 15)(r16 16)(r17 17)(r18 18)(r19 19)(r20 20)
	   (r21 21)(r22 22)(r23 23)(r24 24)(r25 25)(r26 26)
	   (r27 27)(r28 28)(r29 29)(r30 30)(r31 31))
  '*register-code)

%%
%%	Psl friendly names of registers.
%%

(deflist
  '((argone r3) (argtwo r4) (argthree r5) (argfour r6) (argfive r7)
    (t1 r19) (t2 r8) (t3 r9) (t4 r10) (t5 r11) (t6 r12)
    (NIL r13) (st r1) (sp r1) (unix-st r1) (returnaddr r31)
    (mq 0) (xer 1) (lr 8) (ctr 9)
    (bndstkptr r14) (bndstklowerbound r15) (bndstkupperbound r16)
    (symfnccenter r17) (symvalcenter r18)) '*other-name)

(put 'nil '*other-name '13) %doesnt work in deflist ??
(put 't   '*other-name '0)

(deflist
  '(    (heaplast r20) (heaptrapbound r21) (t r0)
        (bndstkptr r14) (bndstklowerbound r15) (bndstkupperbound r16)
        ) 'Registername)
))

% Registers 22-30 are used for frame (9 elements)
% Link Register is copied into R31 after the frame is dumped to memory
% this way it is saved by the next callee.

(de expand-operand (asm-code)
  (cond ((atom asm-code) asm-code)
        (t   (cons (car asm-code)
               (mapcar (cdr asm-code)
                       (function (lambda(x)(to-standard-name x))))))))

(global '(*REGISTER-NAMES*))
 
(setq *REGISTER-NAMES*  '((0 0) (1 3)(2 4)(3 5)(4 6)(5 7)))
 
(de to-standard-name (u)
  (cond ((idp u) (cond ((get u '*other-name) (get u '*other-name))
                        (t u)))
        ((pairp u)
         (cond ((eq (car u) 'REG)
                    (cond ((numberp (cadr u)) 
			   (if  (wlessp (cadr u) 6)
                            (cadr (assoc (cadr u) *REGISTER-NAMES*)) (cadr u)))
                         ((idp (cadr u)) (to-standard-name (cadr u)))
                         (t (error 0 (list "Unknown RS6000 register name" u)))))
               ((member (car u) '(MKITEM WARRAY)) u)
               ((eq (car u) 'DEFERRED)   (list (to-standard-name (cadr u))))
               ((member (car u) '(INDEXED DISPLACEMENT))
                (list 'displacement (to-standard-name (cadr u))
                      		    (to-standard-name (caddr u))))
               ((eq (car u) '+) (cadr u))
               ((eq (car u) '-) u)
	       ((memq (car u) '($fluid $global)) u)
               (t (mapcar u (function to-standard-name)))))
        (t u)))
 
(de ibmrs-register-p (r)
  (cond ((idp r)
         (cond ((get r 'Registername) T)
               ((get r '*other-name) (ibmrs-register-p (get r '*other-name)))
               ((get r '*register-code) t)
               ((get r '*f-register-code) t)
               ((get r '*c-register-code) t)
               (t nil)))
        ((eqcar r 'REG) t)
        (t nil)))
 
(de get-register-code (reg)
      (cond ((pairp reg) (get-register-code (cadr reg)))
            ((and (get reg '*other-name)
                        (get-register-code (get reg '*other-name))))
            ((eq reg 0) 0)
            ((and (numberp reg) (lessp reg 6)) (plus2 reg 2))
	    ((numberp reg) reg)
            ((get reg '*register-code) (get reg '*register-code))
            (t (error 0 (list "Wrong register argument" reg))))
)
              
(de regp (r) (eqcar r 'reg))

(de fregp (r) (regp r))
