%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC:SPARC-CMAC.sl
% Title:          Patterns and predicates for Sparc PSL cmacro expansion
% Author:         Winfried Neun, using Chris Burdorf's sparc-cmacros
% Created:        December 15, 1988
% Modified:
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions
%  2. Jan 1990 (Winfried Neun)
%  Included immediatequote stuff
%  20 April 1989 (Winfried Neun)
%  Cleaning up, better car and cdr expansions
%  30 Jan 1989 (Winfried Neun)
%  Optimized loading of items from memory addresses in code (*movex)
%  Better *JUMPON now, using jmpl instruction
%  Revised alloc scheme in  order to make stack frames self explanatory for
%  gc and backtrace

(on fast-integers)

(compiletime
  (progn (on eolinstringok)
     (dm * (u)  nil)
     (setq *syslisp t)
     (load if-system)
     nil))

(fluid '(AddressingUnitsPerItem
     CharactersPerWord
     StackDirection
     *ImmediateQuote
     AddressingUnitsPerFunctionCell))

(loadtime (on ImmediateQuote))

(flag '(Immediate UnImmediate Indirect displacement predecrement
	SethiIdloc
    label extrareg postincrement Indexed absolute deferred reglist)
    'TerminalOperand)

% MkItem may be used when evaluating WConst expressions.

(de MkItem (TagPart InfPart) % ZIB
  (lor (lsh TagPart 27) (land InfPart 16#7ffffff)))

(de freep (asize) (and (eq 0 asize) (not &links-present&)))

(de FiveP (x)  (eq x 5))

(de TwentySevenP (x)  (eq x 27))

(de compiler-evenp (x) (and (fixp x) (eq 0 (land x 1))))

(de indexedp (x) (eqcar x 'indexed))

(de DeferrableP (Expression)
  (and (PairP Expression)
       (not (memq (first Expression)
              '(Deferred $Fluid $Global Fluid Global)))))

(de RegisterDeferredP (Expression)
  (and (EqCar Expression 'Deferred) (EqCar (second Expression) 'Reg)))


(de FakeRegisterNumberP (Expression)
  (and (IntP Expression) (GreaterP Expression 5)))

% NegativeImmediateP is for numbers whose negative is a one byte operand.

(de negp (value) (and (numberp value) (lessp value 0)))

(de displacementp (x) (and (pairp x) (eq (car x) 'displacement)))

(de NegativeImmediateP (Expression)
  (and (FixP Expression) (minusp Expression) (GreaterP Expression -64)))

% expansions of addressing modes.

(DefAnyreg CAR  ZIBanyregCAR)  % ZIB

(de ZIBAnyregCAR (reg source)
   (prog (areg)
      (setq source (resolveoperand reg source))
      (if (and (displacementp source)     % recursive case, e.g. caar
               (member (cadr source) *nextreglist*))
             (setq areg (cadr source))
             (setq areg (nexttempreg)))
      (if (Xregp source)  % its a reg or a frame reg
          (push `(andn ,source (reg i5) ,areg) Resultingcode*)

          (push `(*move ,source , areg) Resultingcode*)
          (push `(andn ,areg (reg i5) ,areg) Resultingcode*))
      (return `(displacement ,areg 0))))

(DefAnyreg CDR  ZIBanyregCDR)  % ZIB

(de ZIBAnyregCDR (reg source)
   (prog (areg)
      (setq source (resolveoperand reg source))
      (if (and (displacementp source)     % recursive case, e.g. cddar
               (member (cadr source) *nextreglist*))
             (setq areg (cadr source))
             (setq areg (nexttempreg)))
      (if (Xregp source)  % its a reg or a frame reg
          (push `(andn ,source (reg i5) ,areg) Resultingcode*)

          (push `(*move ,source , areg) Resultingcode*)
          (push `(andn ,areg (reg i5) ,areg) Resultingcode*))
      (return `(displacement ,areg 4))))

(fluid '(*nextreg* *nextreglist*))

(setq *nextreg* 0        % Four registers should be sufficient
      *nextreglist* '((reg t2) (reg t3) (reg t4) ))

(de nexttempreg (x)   % ZIB  :This saves Regs , non static assignment
     (setq *nextreg* (wplus2 *nextreg* 1))
     (when (eq *nextreg* 4) (setq *nextreg* 1))
     (nth *nextreglist* *nextreg*))

(DefAnyreg Memory  ZIBanyregMemory)  % ZIB

(de ZIBAnyregMemory (reg source offset)
  (prog (aReg cadrsource)
   (setq source (resolveoperand reg source))
   (setq offset (resolveoperand reg offset))
   (cond ((and (registerp source) (registerp offset))
             (return `(indexed ,source ,offset))) % SUN 4 has double index
         ((registerp source)
      (return (if (eq 0 offset) `(Deferred ,source)
                     (if (sparc-short-immediate-p offset)
                            `(displacement ,source ,offset)
                            (progn
                              (setq aReg (nexttempreg))
                                             % take a new TempReg
                              (push `(*move ,offset ,aReg) Resultingcode*)
                             `(displacement ,source ,aReg) )))))
    % what shall that ?
      ((and (AddressConstantP source)(ZeroP offset))
	    (return  `(unimmediate ,SOURCE )))
%%%%%%%   ((and (DeferrableP source) (ZeroP offset))
%%%%%%%    (return `(Deferred ,SOURCE) ))
      ((and (RegisterDeferredP source )(ZeroP offset))
         (return `(deferred (Displacement (UnDeferred ,SOURCE) 0)))))
    (setq aReg (if (and (displacementp source)
                    (member (setq cadrsource (cadr source))
                                 *nextreglist*))
               cadrsource      (nexttempreg))) % take a new TempReg
    (push `(*move ,source ,aReg) Resultingcode*)
    (return `(displacement ,aReg  ,offset))))


(DefAnyreg QUOTE
       AnyregQUOTE
       ((Null) (REG NIL))
       ((EqTP) (indexed (reg g7) -36)) %%FLUID T))
       ((InumP) SOURCE)
       (       (QUOTE SOURCE)))

(fluid '(&frame-regs&))

(setq &frame-regs& [ nil l1 l2 l3 l4 l5 l6 l7 i0 i1 i2 i3 i4 ])

(DefAnyreg FRAME ZIBAnyregFrame)  % ZIB

(de ZIBAnyregFRAME (dum number)
    (cond ((wgeq number 13)
             `(displacement (reg st) ,(plus2 48 (times2 4 number))))
      (t `(reg ,(indx &Frame-regs& number)))))

(DefAnyreg REG ZIBAnyregREG)

(de ZIBanyregReg (reg source)
   (cond ((Fakeregisternumberp source)
          `(indexed (reg g7) ,(times2 4 (plus2 source -6))))
     ((and (idp source) (get source 'viv-number)) % fast access var?
          `(indexed (reg g7) ,(get source 'viv-number)))
     (t       `(REG ,SOURCE))))

(loadtime (remflag '(fluid global $fluid $global) 'Terminaloperand))

(defanyreg Fluid ZIBanyregfluid)

(defanyreg $Fluid ZIBanyregfluid)

(fluid '(*syslisp ))

(de ZIBanyregFluid (reg source)
   (cond ((and (idp source) (get source 'viv-number) % fast access var?
         ) % (or *syslisp (flagp source '*user-variable))) % safety !!
                `(indexed (reg g7) ,(get source 'viv-number)))
     (t       `($Fluid ,SOURCE))))

(defanyreg $VivFluid ZIBanyregVivfluid)  % for usage for VIV in non syslisp
                                     % e.g. lambind, cons ... etc
(de ZIBanyregVIVFluid (reg source)
  (let ((getviv (get source 'viv-number)))
     (cond (getviv  `(indexed (reg g7) ,getviv))
       (t       `(*Cerror " Wrong VivFluid" ,SOURCE)))))

(defanyreg Global ZIBanyregglobal)

(defanyreg $Global ZIBanyregglobal)

(de ZIBanyregGlobal (reg source)
   (cond ((and (idp source) (get source 'viv-number) % fast access var?
           )% (or *syslisp (flagp source '*user-variable))) % safety !!
                `(indexed (reg g7) ,(get source 'viv-number)))
     (t       `($global ,SOURCE))))

%-----------------------------------------------------------------------------

(de *LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
  (if (FlagP FunctionName 'ForeignFunction)
      (list (list '*ForeignLink FunctionName FunctionType NumberOfArguments)
        '(*Exit 0))
      (list (list '*DeAlloc DeAllocCount)
        (list '*JCall FunctionName))))

(DefCMacro *Call
       ((InternallyCallableP)  (call (InternalEntry ARGONE))
                           (NOP))
        ( (*move  ($fluid symfnc) (reg t3))
          (*Move (sethiidloc ARGONE) (reg t1))
	  (srl (reg t1) 8 (reg t4))
	  (*move (displacement (reg t3) (reg t4)) (reg t4))
	  (jmpl (displacement (reg t4) 0) (reg 15))
          (nop)))
 
(DefCMacro *JCall
       ((InternallyCallableP) (BA (InternalEntry ARGONE))
                            (nop))
        ( (*move  ($fluid symfnc) (reg t3))
          (*Move (sethiidloc ARGONE) (reg t1))
          (srl (reg t1) 8 (reg t4))
          (*move (displacement (reg t3) (reg t4)) (reg t4))
          (jmpl (displacement (reg t4) 0) (reg 0))
          (nop)))

(de quote-fix-p (x)
    (and (eqcar x 'quote) (fixp (cadr x))))

(de displacement-p (x)
    (and (eqcar x 'displacement)
     (eqn (length x) 3) (sparc-register-p (cadr x))))

(de deferred-p (x)
    (and (pairp x) (or (sparc-register-p (car x))
                   (eq (car x) 'deferred))))

(remprop '*move 'optfn)

(DefCMacro *Move
       (Equal)
       ((Xregp Xregp)               (mov argone argtwo))
       ((schreckenshugop Xregp)     (*schrecken argone argtwo))
       ((indexedp Xregp)            (ld argone argtwo))
       ((Xregp indexedp)            (st argone argtwo))
       ((Xregp anyp)                (*movex argone argtwo))
       ((anyp Xregp)                (*movex argone argtwo))
       (                            (*move argone (reg t5))
                                    (*move (reg t5) Argtwo)))

% for gc: put a hskip item  into frame in correct length

(de schreckenshugop (x) (and (eqcar x 'immediate)
                         (pairp (cadr x))
                         (eqcar (cadr x) 'schreckenshugo)))

(de *schrecken (argone argtwo)
    (setq argone (cadr (cadr argone))) % (immediate (schreckenshugo n)
    `((sethi (hi (lshift ,(wor 16#360000 argone)10)) ,argtwo)))
		 % its a forward tag

(defcmacro *schrecken)

(de *ALLOC (framesize)

    (setq NAlloc!* framesize)
 % use standard allocation scheme % ZIB
 % we have to leave 7 words unused for foreigncall!! see stack scheme in
 % Sparc architecture manual
    (cond ((freep) NIL)
      ((wgeq framesize 13)
        (if (compiler-evenp framesize)
        `((save (reg st) ,(minus (times 4 (plus2 14 framesize))) (reg st))
              (sethi (hi (lshift 16#3E0000 10)) (reg i5)) % frame 13
          (*move (schreckenshugo ,framesize) (reg l0)))
        `((save (reg st) ,(minus (times 4 (plus2 15 framesize))) (reg st))
              (sethi (hi (lshift 16#3E0000 10)) (reg i5)) % frame 13
          (*move (schreckenshugo ,framesize) (reg l0)))
       )) % must be a multiple of 8 !!
      (t `((save (reg st) ,-96  (reg st))
              (sethi (hi (lshift 16#3E0000 10)) (reg i5)) % frame 13
          (*move (schreckenshugo ,framesize) (reg l0))))
)   )

(defcmacro *ALLOC)

(DefCMacro *DeAlloc
       ((FreeP))
       (         (restore  (reg 0) ARGONE (reg 0))))

(DefCMacro *Exit
     % leaf routine first
       ((Freep)  (jmpl (displacement (reg 15) 8) (reg 0))
                 (nop))
       (         (jmpl (displacement (reg returnaddr) 8) (reg 0))
                 (restore (reg 0) 0 (reg 0))))

(de deferredp (x) (and (pairp x) (eq (car x) 'deferred)))

(DefCMacro *Jump
       ((deferredp)        (*move (undeferred ARGONE) (reg t5))
                           (jmpl (deferred (reg t5)) (reg 0))
                               (or (reg 0) 0 (reg 0))) % NO NOP!!!!! FOR LAPOPT
       ((displacementp)    (jmpl ARGONE (reg 0))
                           (nop))
       (                   (ba ARGONE)
                           (nop)))


(DefCMacro *Lbl           (ARGONE))

(de Xregp (x) (or (eqcar x 'Reg)
              (and (eqcar x 'frame) (leq (cadr x) 12))))

(de *3op (a1 a2 instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) A1)
                 (ResolveOperand '(Reg t2) A2)
                  Instruction)
                 (get '*3op 'CMacroPatternTable)))))

(DefCMacro *3op
       ((Xregp Xregp) (ARGTHREE ARGOne ARGtwo ARGONE))
       ((Xregp sparc-short-immediate-p)
                      (ARGTHREE ARGONE ARGTWO ARGONE))
       ((Xregp AnyP)  (*move argtwo (reg T5))
                      (*3op argone (reg t5) argthree))
       ((Anyp Xregp)  (*move argone (reg T5))
                      (ARGTHREE (reg t5) argtwo (reg t5))
                      (*move (reg t5) argone))
       ((Anyp sparc-short-immediate-p)
                      (*move argone (reg T6))
                      (argthree (reg t6) argtwo (reg t5))
                      (*move (reg t5) argone))
       (              (*move argtwo (reg T5))
                      (*move argone (reg T6))
                      (argthree (reg t6) (reg t5) (reg t5))
                      (*move (reg t5) argone)))

(DefCMacro *WPlus2
       ((*3op ARGOne Argtwo Add)))

(de *wplus3(arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wplus3))

(DefCMacro *WPlus3
           ((xregp xregp xregp) (add argthree argtwo argone))
           ((xregp sparc-short-immediate-p xregp)
                                (add argthree argtwo argone))
           ((xregp xregp sparc-short-immediate-p)
                                (add argtwo argthree argone))
           ((xregp sparc-short-immediate-p anyp)
                               (*move argthree (reg t5))
                               (add (reg t5) argtwo argone))
           ((xregp anyp sparc-short-immediate-p)
                               (*move argtwo (reg t5))
                               (add (reg t5) argthree argone))
           ((xregp anyp anyp)  (*move argtwo argone)
                               (*wplus2 argone argthree))
           ((anyp xregp anyp)  (*move argthree (reg t5))
                               (*wplus2 (reg t5) argtwo)
                               (*move (reg t5) argone))
           (                   (*move argtwo (reg t5))
                               (*move argthree (reg t6))
                               (add (reg t5) (reg t6) (reg t5))
                               (*move (reg t5) argone)))

(DefCMacro *WDifference
       ((*3op ARGONE ARGTWO Sub)))

(de *wdifference3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wdifference3))

% format argone = argtwo - argthree

(DefCMacro *Wdifference3
           ((xregp xregp xregp) (sub argtwo argthree argone))
           ((xregp xregp sparc-short-immediate-p)
                               (sub argtwo argthree argone))
           ((xregp anyp xregp) (*move argtwo (reg t5))
                               (sub (reg t5) argthree argone))
           ((xregp anyp anyp)  (*move argtwo (reg t5))
                               (*move argthree (reg t6))
                               (sub (reg t5) (reg t6) argone))
           ((anyp xregp anyp)  (*move argthree (reg t5))
                               (sub  argtwo (reg t5) (reg t5))
                               (*move (reg t5) argone))
           (                   (*move argtwo (reg t5))
                               (*move argthree (reg t6))
                               (sub (reg t5) (reg t6) (reg t5))
                               (*move (reg t5) argone)))

(defcmacro *wtimes2)

(put 'wtimes2 'opencode '((smul (reg 1) (reg 2) (reg 1))))
(put 'wquotient 'opencode '((sra (reg 1) 31 (reg 4))
                            (wry (reg 4) (reg 0) (reg 0))
                            (sdiv (reg 1) (reg 2) (reg 1))))
(put 'wremainder 'opencode '((sra (reg 1) 31 (reg 4))
			    (wry (reg 4) (reg 0) (reg 0))
		            (SDIV (reg 1) (reg 2) (reg 3))
% quotient is in REG 3
		            (smul (reg 3) (reg 2) (reg 4))
        		    (*wdifference (reg 1) (reg 4))))

(de *wtimes3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wtimes3))

% format argone = argtwo * argthree

(DefCMacro *wtimes3
           (        (*move argtwo   ($fluid arithargs0))
                    (*move argthree ($fluid arithargs1))
                    (*move ($fluid arithargs0) (reg f0))
                    (*move ($fluid arithargs1) (reg f1))
                    (fitod (reg f0) (reg f2))
                    (fitod (reg f1) (reg f4))
                    (fmuld (reg f2) (reg f4) (reg f6))
                    (fdtoi (reg f6) (reg f7))
                    (st (reg f7) (indexed (reg st) 72)) % this is unused
                    (*move (indexed (reg st) 72) argone)))

(defcmacro *wquotient)

(de *wquotient (arg1 arg2) (*wquotient3 arg1 arg1 arg2))

(de *wquotient3 (arg1 arg2 arg3)
    (expand3operandcmacro arg1 arg2 arg3 '*wquotient3))

% format argone = argtwo * argthree

(DefCMacro *wquotient3
	   ((anyp anyp onep)  (*move argtwo argone))
	   ((xregp fixp fixp)  (*move (quotient ArgTwo ArgThree) ArgOne))
           ((xregp xregp powerof2p)
                    (*wquotientshift argone argtwo argthree))
           (        (*move argtwo   ($fluid arithargs0))
                    (*move argthree ($fluid arithargs1))
                    (*move ($fluid arithargs0) (reg f0))
                    (*move ($fluid arithargs1) (reg f1))
                    (fitod (reg f0) (reg f2))
                    (fitod (reg f1) (reg f4))
                    (fdivd (reg f2) (reg f4) (reg f6))
                    (fdtoi (reg f6) (reg f7))
                    (st (reg f7) (indexed (reg st) 72)) % this is unused
                    (*move (indexed (reg st) 72) argone)))

(de *wquotientshift (arg1 arg2 arg3)
        (setq arg3 (powerof2p arg3))
        `((srl ,arg2 ,(difference 32 arg3) (reg t5))
          (*wplus2 (reg t5) ,arg2)
          (sra (reg t5) ,arg3 ,arg1)))

(defcmacro *wquotientshift)

(DefCMacro *WAnd
       ((*3op ARGONE ARGTWO and)))

(DefCMacro *WOr
       ((*3op ARGONE ARGTWO or)))

(DefCMacro *WXOr
       ((*3op ARGONE ARGTWO xor)))

(de fixplusp (x) (and (fixp x) (not (minusp x))))

(DefCMacro *AShift
    ((AnyP ZeroP))
    ((Xregp negp)     (sra ARGONE (minus ARGTWO) ARGONE))
    ((Xregp fixplusp) (sll ARGONE ARGTWO ARGONE))
    ((Xregp Xregp)    (subcc ARGTWO (reg 0) (reg 0))
                      (bl!,a templabel)
                      (sub (reg 0) ARGTWO (reg t6)) % delay slot
                      (sll ARGONE ARGTWO ARGONE)
                      (ba templabel2)
                      (nop)
                     templabel
                      (sra ARGONE (reg t6) ARGONE)
                     templabel2)
    ((Xregp anyp)     (*move argtwo (reg t5))
                      (*ashift argone (reg t5)))
    ((Anyp fixp)      (*move argone (reg t5))
                      (*ashift (reg t5) argtwo)
		      (*move (reg t5) argone))
    (                 (*move argone (reg t1))
                      (*ashift (reg t1) argtwo)
                      (*move (reg t1) argone)))


(DefCMacro *WShift
        ((AnyP ZeroP))
    ((Xregp negp)     (srl ARGONE (minus ARGTWO) ARGONE))
    ((Xregp fixplusp) (sll ARGONE ARGTWO ARGONE))
    ((Xregp Xregp)    (subcc ARGTWO (reg 0) (reg 0))
                      (bl!,a templabel)
                      (sub (reg 0) ARGTWO (reg t6)) % delay slot
                      (sll ARGONE ARGTWO ARGONE)
                      (ba templabel2)
                      (nop)
                     templabel
                      (srl ARGONE (reg t6) ARGONE)
                     templabel2)
    ((Xregp anyp)     (*move argtwo (reg t5))
                      (*wshift argone (reg t5)))
    ((Anyp fixp)      (*move argone (reg t5))
                      (*wshift (reg t5) argtwo)
		      (*move (reg t5) argone))
    (                 (*move argone (reg t1))
                      (*wshift (reg t1) argtwo)
                      (*move (reg t1) argone)))

(DefCMacro *WNot
       ((Xregp Xregp) (Xnor ARGTWO (reg g0) ARGONE))
       ((Xregp Anyp)  (*move argtwo argone)
                      (Xnor argone (reg g0) argone))
       (              (*move argtwo (reg t5))
                      (*wnot (reg t5) (reg t5))
                      (*move (reg t5) argone)))

(DefCMacro *WMinus
           ((Xregp Xregp) (neg ARGTWO ARGONE))
           ((Xregp Anyp)  (*move argtwo argone)
                          (neg argone argone))
           (              (*move argtwo (reg t5))
                          (*wminus (reg t5) (reg t5))
                          (*move (reg t5) argone)))

(DefCMacro *MkItem
       ((Xregp fixp)      (andn argone (reg i5) argone)
                          (*specsethi argtwo)      % sethi with tag to t5
                          (or argone (reg t5) argone))
       ((Xregp Xregp)     (andn argone (reg i5) argone)
                          (sll argtwo 27 (reg t5))
                          (or argone (reg t5) argone))
       ((Xregp anyp)      (*move argtwo (reg t5))
                          (*Mkitem argone (reg t5)))
       (                  (*move argone (reg t1))
                          (*mkitem (reg t1) argtwo)
                          (*move (reg t1) argone)))

(de *specsethi (num)          % avoid shift of constant at compiletime
      (if (or !*writingasmfile (getenv "linux"))
        `((sethi (hi ,(wshift num 17)) (reg t5))
	  (sll (reg t5) 10 (reg t5)))
        %% why this  ??`((sethi (hi ,(wshift num 27)) (reg t5)))))
        `((sethi ,(wshift num 17) (reg t5)))))

(defcmacro *specsethi)  % support for *mkitem

(de smallinump (x) (sparc-short-immediate-p x))

(de immediate-p (x)    (fixp x))

(DefCmacro *tag
      ((Xregp Xregp)(srl argtwo 27 argone))
      ((Anyp  Xregp)(*move argtwo (reg t1))
                    (srl (reg t1) 27 (reg t1))
                    (*move (reg t1) argone))
      ((xregp anyp) (*move argtwo (reg t1))
                    (srl (reg t1) 27 argone))
      (             (*move argtwo (reg t1))
                    (*tag argone (reg t1))))

(de *tag (ar1 ar2)
     (expand2operandCmacro ar1 ar2 '*tag))

(DefCMacro *JumpType
      (             (*Tag (Reg t5) ArgOne)
                    (*JumpEQ ArgThree (Reg t5) ArgTwo)))

(DefCMacro *JumpNotType
      (             (*Tag (Reg t5) ArgOne)
                    (*JumpNOTEQ ArgThree (Reg t5) ArgTwo)))

(DefCMacro *JumpInType
       ((Xregp Anyp) (srl ARGONE 27 (reg t5))
                     (subcc (reg t5) ARGTWO (reg 0))
                     (ble ARGTHREE)
                     (subcc (reg t5) 31 (reg 0)) % negint-tag)
                     (be ARGTHREE)
                     (nop))
       (             (*move argone (reg t5))
                     (srl (Reg t5) 27 (reg t5))
                     (subcc (reg t5) ARGTWO (reg 0))
                     (ble ARGTHREE)
                     (subcc (reg t5) 31 (reg 0)) % negint-tag)
                     (be ARGTHREE)
                     (nop)))

(DefCMacro *JumpNotInType
       ((Xregp Anyp) (srl ARGONE 27 (reg t5))
                     (subcc (reg t5) 31 (reg 0))
                     (be templabel)
                     (subcc (reg t5) argtwo (reg 0))
                     (bg ARGTHREE)
                     (nop)
                    templabel)
       (             (*move argone (reg t5))
                     (srl (reg t5) 27 (reg t5))
                     (subcc (reg t5) 31 (reg 0))
                     (be templabel)
                     (subcc (reg t5) argtwo (reg 0))
                     (bg ARGTHREE)
                     (nop)
                    templabel))


%---------------------------------------------------------------------
%   General Purpose Jumper.
%   FORMAT: Operand1 Operand2 Address JumpOpCode

(DefCMacro *JumpIf
     ((Xregp Xregp)      (subcc argone argtwo (reg g0))
                         (argfour argthree)
                         (nop))
     ((XRegP SmallInumP) (subcc argone argtwo (reg g0))
                         (argfour argthree)
                         (nop))
     ((Xregp anyp)       (*move argtwo (reg t5))
                         (subcc argone (reg t5) (reg g0))
                         (argfour argthree)
                         (nop))
     ((anyp  Xregp)      (*move argone (reg t5))
                         (subcc (reg t5) argtwo  (reg g0))
                         (argfour argthree)
                         (nop))
     ((anyp  SmallInumP) (*move argone (reg t5))
                         (subcc (reg t5) argtwo  (reg g0))
                         (argfour argthree)
                         (nop))
     (                   (*move argone (reg t5))
                         (*move argtwo (reg t4))
                         (subcc (reg t5) (reg t4) (reg g0))
                         (argfour argthree)
                         (nop))
)

(de *JumpIF (ArgOne ArgTwo Label Instruction)
  (prog (ResultingCode*)
    (return (CMacroPatternExpand
           (list (ResolveOperand '(Reg t1) ArgOne)
                 (ResolveOperand '(Reg t2) ArgTwo)
                     (ResolveOperand '(Reg Error) Label)
                  Instruction)
                 (get '*JumpIF 'CMacroPatternTable)))))


(defcmacro *jumpeq)

(de *jumpeq (Lbl ArgOne ArgTwo)
       (*jumpif argone argtwo lbl 'BE!,a))

(defcmacro *jumpnoteq)

(de *jumpnoteq(Lbl ArgOne ArgTwo)
        (*jumpif argone argtwo Lbl 'BNE!,a))

(defcmacro *jumpwlessp)

(de *jumpwlessp (Lbl ArgOne ArgTwo)
        (*jumpif argone argtwo lbl 'BL!,a))

(defcmacro *jumpwgreaterp)

(de *jumpwgreaterp (Lbl ArgOne ArgTwo)
        (*jumpif argone argtwo lbl 'BG!,a))

(defcmacro *jumpwleq)

(de  *jumpwleq(Lbl ArgOne ArgTwo)
        (*jumpif argone argtwo lbl 'BLE!,a))

(defcmacro *jumpwgeq)

(de *jumpwgeq (Lbl ArgOne ArgTwo)
        (*jumpif argone argtwo lbl 'BGE!,a))


(de fluid-arg-p (x)
    (and (pairp x)
     (member (car x) '(!$fluid !$global fluid global))))


(DefCMacro *Push
       ( (*cerror "NO push ")))

(DefCMacro *Pop
       ( (*cerror "NO pop ")))

(de take-car (x) (car x))
(de take-cadr (x) (cadr x))   % functions to strip out deferred and
                          % displacement parts.
(de take-caddr (x) (caddr x))


(de *loc (argone argtwo)
   (let ((resolvedarg (resolveoperand '(reg t2) argtwo)))
      (cond ((deferred-p resolvedarg)
         `((*move (undeferred ,resolvedarg) ,argone))) % get the address.
         ((displacement-p resolvedarg)
            `((add ,(cadr resolvedarg) ,(caddr resolvedarg) ,argone)))
         ((eqcar resolvedarg 'cdr)
            `((add ,(cadr resolvedarg) 4 ,argone)))
         (t `((*movex ,resolvedarg ,argone))))))

(DefCMacro *loc)

%t6 is used, because it is not used in lap code, but it destroyed in cmacros.
%look at test file mask.s in ~burdorf/sparc for example.

(de X5p (u) (eq u 5))

(de X27p (u) (eq u 27))

(DefCMacro *Field

  ((Xregp Xregp zerop fixp)  (srl ARGTWO (difference 32 ARGFOUR) ARGONE))
  ((Xregp Xregp X5p X27p)    (andn ARGTWO (reg i5) Argone))
  ((Xregp Xregp fixp fixp)   (sll argtwo argthree argone)
                         (srl argone (difference 32 ARGFOUR) ARGONE))
  ((Xregp anyp fixp fixp)    (*move argtwo (reg t5))
                         (sll (reg t5) argthree (reg t5))
                         (srl (reg t5) (difference 32 ARGFOUR) ARGONE))
  ((anyp Xregp fixp fixp)    (*field (reg t5) argtwo argthree argfour)
                         (*move (reg t5) argone))
  (                          (*move argtwo (reg t6))
                         (*field (reg t5) (reg t6) argthree argfour)
                         (*move (reg t5) argone)))

(DefCMacro *SignedField

  ((Xregp Xregp fixp fixp)   (sll argtwo argthree argone)
                         (sra argone (difference 32 ARGFOUR) ARGONE))
  ((Xregp anyp fixp fixp)    (*move argtwo (reg t5))
                         (sll (reg t5) argthree (reg t5))
                         (sra (reg t5) (difference 32 ARGFOUR) ARGONE))
  ((anyp Xregp fixp fixp)    (*Signedfield (reg t5) argtwo argthree argfour)
                         (*move (reg t5) argone))
  (                          (*move argtwo (reg t6))
                         (*Signedfield (reg t5) (reg t6) argthree argfour)
                         (*move (reg t5) argone)))

% following is used by garbage collector only. Not very useful to
% optimize other cases than put inf field or put tag field (with constant)


(DefCMacro *PutField

 ((Xregp XregP fiveP TwentysevenP) (andn argone (reg i5) (reg t6))
                                   (and   (reg i5) argtwo argtwo)
                                   (or    argtwo (reg t6) ARGTWO))
 ((Xregp Xregp Zerop fivep)        (sll   argone 27 (reg t5))
                               (andn  argtwo (reg i5) argtwo)
                               (or    (reg t5) argtwo argtwo))
 ((inump Xregp Zerop fivep)        (mov   argone (reg t5))
                               (sll   (reg t5) 27 (reg t5))
                               (andn  argtwo (reg i5) argtwo)
                               (or    (reg t5) argtwo argtwo))
 (          (mov -1  (reg t5))
        (*move   argone (reg t4))
        (sll     (reg t5) (difference 32 argfour) (reg t5))
        (*wshift (reg t5) (minus argthree)) % for 0 opt
        (sll   (reg t4) (difference 32 (plus argthree argfour)) (reg t4))
        (and   (reg t5) (reg t4) (reg t4))
        (*move argtwo (reg t6))
        (andn  (reg t6) (reg t5) (reg t6))
        (or   (reg t4) (reg t6) (reg t5))
        (*move (reg t5) argtwo)))


(deflist '((Byte        ((ldsb (displacement (reg 2) (reg 1)) (reg 1))))
       (PutByte     ((stb (reg 3) ((reg 1) (reg 2)))))
       (HalfWord    ((sll (reg 2) 1 (reg t1))
                     (ldsh (displacement (reg t1)  (reg 1)) (reg 1))))
       (PutHalfWord ((sll (reg 2) 1 (reg t1))
                     (sth (reg 3) (displacement (reg 1) (reg t1))))))
  'OpenCode)

(&OneReg '(Byte PutByte HalfWord PutHalfWord))

(on fast-integers)

(for (in function '(idapply0 idapply1 idapply2 idapply3 idapply4))
     (from register 1)
     (do
      (put function 'opencode
        `((*move ($fluid symfnc) (reg t1))
          (sll (reg ,register) 2 (reg ,register))
          (ld (displacement (reg t2) (reg ,register)) (reg t2))
          (jmpl (displacement (reg t2) 0) (reg 15))
              (sll (reg ,register) 8 (reg t1)))
      (put function 'exitopencode
        `((*move ($fluid symfnc) (reg t1))
          (sll (reg ,register) 2 (reg ,register))
          (ld (displacement (reg t2) (reg ,register)) (reg t2))
          (jmpl (displacement (reg t2) 0) (reg 0))
              (sll (reg ,register) 8 (reg t1)))   
   ))))

(for (in function '(addressapply0 addressapply1 addressapply2
                addressapply3 addressapply4))
     (from register 1)
     (do
      (put function 'opencode
         `((jmpl (deferred (reg ,register)) (reg 15))
               (or (reg 0) 0 (reg 0)))) % NO NOP!!!!! FOR LAPOPT
      (put function 'exitopencode
         `((jmpl (deferred (reg ,register)) (reg 0))
             (or (reg 0) 0 (reg 0)))) % NO NOP!!!!! FOR LAPOPT
      ))


% *feq, *fgreaterp and *flessp can only occur once in a function.

(deflist '((*wfix ((ldd (deferred (reg 1)) (reg f0))
               (fdtoi (reg f0) (reg f2))
               (st  (reg f2) (displacement (reg sp)  -4))
               (ld  (displacement  (reg sp) -4) (reg 1))))
       (*wfloat ((st (reg 2) (displacement (reg sp) -4))
                 (ld (displacement (reg sp) -4) (reg f0))
                 (fitod (reg f0) (reg f2))
                 (std (reg f2) (deferred (reg 1)))))
       (*feq ((ldd  (deferred (reg 1)) (reg f0))
              (ldd (deferred (reg 2)) (reg f2))
              (fcmpd (reg f0) (reg f2))
              (fbe  *donefeq*)
              (nop)
              (mov (reg nil) (reg 1))
              *donefeq*))
       (*fgreaterp ((ldd  (deferred (reg 1)) (reg f0))
                    (ldd (deferred (reg 2)) (reg f2))
                    (fcmpd (reg f0) (reg f2))
                        (*move (quote t) (reg 1))
                    (fbg  *donefgreaterp*)
                    (nop)
                    (mov (reg nil) (reg 1))
                    *donefgreaterp*))
       (*flessp ((ldd  (deferred (reg 1)) (reg f0))
                 (ldd (deferred (reg 2)) (reg f2))
                 (fcmpd (reg f0) (reg f2))
                     (*move (quote t) (reg 1))
                 (fbl *doneflessp*)
                 (nop)
                 (mov (reg nil) (reg 1))
                 *doneflessp*))
       (*fplus2 ((ldd  (deferred (reg 2)) (reg f0))
                 (ldd (deferred (reg 3)) (reg f2))
                 (faddd (reg f0) (reg f2) (reg f0))
                 (std (reg f0) (deferred (reg 1)))))
       (*fdifference ((ldd  (deferred (reg 3)) (reg f0))
                      (ldd (deferred (reg 2)) (reg f2))
                      (fsubd (reg f2) (reg f0) (reg f0))
                      (std (reg f0) (deferred (reg 1)))))
       (*ftimes2 ((ldd  (deferred (reg 2)) (reg f0))
                  (ldd (deferred (reg 3)) (reg f2))
                  (fmuld (reg f0) (reg f2) (reg f0))
                  (std (reg f0) (deferred (reg 1)))))
       (*fquotient ((ldd  (deferred (reg 3)) (reg f0))
                    (ldd (deferred (reg 2)) (reg f2))
                    (fdivd  (reg f2) (reg f0) (reg f0))
                    (std (reg f0) (deferred (reg 1))))))
     'opencode)

(defcmacro *fast-apply-load
   ( (*move argone ($fluid *second-value*)) ))

(put 'fast-idapply    'opencode
     '((*move ($fluid symfnc) (reg t5)) %($fluid symfnc) returns address of
                                       %the symfnc table.
       (*move ($fluid *second-value*) (reg t1))
       (sll (reg t1) 5 (reg t1))
       (srl (reg t1) 3 (reg t1))
       (ld (displacement (reg t1) (reg t5)) (reg t5))
       (jmpl (displacement (reg t5) 0) (reg 15))
       (sll (reg t1) 8 (reg t1))))

(put 'fast-codeapply    'opencode
     '((*move ($fluid *second-value*) (reg t1))
       (andn (reg t1) (reg i5) (reg t1))
                            %codepointer passed, so just strip off tag
       (jmpl (deferred (reg t1)) (reg 15))
       (or (reg 0) 0 (reg 0)))) % NO NOP!!!!! FOR LAPOPT

(put 'fast-idapply    'exitopencode
     '((*move ($fluid *second-value*) (reg t1))
       (*move ($fluid symfnc) (reg t5)) %($fluid symfnc) returns address of
                                       %the symfnc table.
       (sll (reg t1) 5 (reg t1))
       (srl (reg t1) 3 (reg t1))
       (ld (displacement (reg t1) (reg t5)) (reg t5))
       (jmpl (displacement (reg t5) 0) (reg 0))
       (sll (reg t1) 8 (reg t1))))

(put 'fast-codeapply    'exitopencode
     '((*move ($fluid *second-value*) (reg t1))
       (andn (reg t1) (reg i5) (reg t1))
                         %codepointer passed, so just strip off tag
       (jmpl (deferred (reg t1)) (reg 0))
       (or (reg 0) 0 (reg 0)))) % NO NOP!!!!! FOR LAPOPT

(De *LamBind (Regs Fluids)

 % be careful, code is somewhat tricky

 (prog (n lng list cregs cfluids cadrcfluids lab initload freeregs)
      (setq lab (gensym))
      (setq n 0)
      (setq Regs (rest Regs))   % Remove "Registers" from the front
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Regs) 8)) % two words per BndStk entry *4

   %looking for free register to preload values of fluid s

      (setq freeregs (mapcan '((reg 1)(reg 2)(reg 3)(reg 4)(reg 5))
                   (function (lambda (x)
                                (cond ((member x regs) nil)
                                       (t (cons x nil))
      )              ) )         )  )
      (setq freeregs (nconc (list '(reg T3) '(reg T4)) freeregs))
      (setq cfluids fluids) % copy of fluids

preload  (setq initload
       (progn (setq cadrcfluids
                (nconc cadrcfluids (cons (car freeregs) nil)))
        (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))

       (cond ((and freeregs cfluids) (go preload)))

       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist

      (setq list `((*move ($VIVfluid BndStkPtr) (Reg t1))
               (*wplus3 (Reg t2) ,lng (Reg t1))
               (*move ($VIVfluid BndstkUpperBound) (Reg t5))
               (subcc (reg t2)(reg t5)(reg g0))
               (bg    (entry Bstackoverflow))
               (*move (Reg t2) ($VIVfluid BndstkPtr))  )) %start of code

      (setq list (append initload list))

  loop
      (setq cregs (car Regs))
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
           (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 8))
      (Setq list (append list
     `((*move ,(car freeregs) (indexed (Reg t1) ,n))
       (*move (quote ,Cadrcfluids)
                   (indexed (reg t1) ,(wplus2 n -4)))
       (*move ,cregs ,cfluids)
      )          ))
      (setq fluids (cdr Fluids))
      (setq freeregs (cdr freeregs))
      (cond ((setq regs (cdr Regs)) (go loop)))
      (return list)
)    )

(defcmacro !*lambind)

(De *ProgBind (Fluids)

 % be careful, code is somewhat tricky

 (prog (n lng list cfluids cadrcfluids lab initload freeregs)
      (setq lab (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 8)) % two words per BndStk entry
                                         % * 4 addressingunits
      (setq freeregs '((reg t3)(reg T4)(reg 1)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids

preload  (setq initload
       (progn (setq cadrcfluids
                (nconc cadrcfluids (cons (car freeregs) nil)))
        (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))

       (cond ((and freeregs cfluids) (go preload)))

       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist


      (setq list `((*move ($VIVfluid BndStkPtr) (Reg t1))
               (*wplus3 (Reg t2) ,lng (Reg t1))
               (*move ($VIVfluid BndstkUpperBound) (Reg t5))
               (subcc (reg t2)(reg t5)(reg g0))
               (bg    (entry Bstackoverflow))
               (*move (Reg t2) ($VIVfluid BndstkPtr))  )) %start of code

     (setq list (append initload list))

 loop
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
           (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 8))
      (Setq list (append list
             `((*move ,(car freeregs) (indexed (Reg t1) ,n))
               (*move (quote ,Cadrcfluids)
                           (indexed (reg t1)  ,(wplus2 n -4)))
               (*move (quote nil) ,cfluids)
      )          ))
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *progbind)

(De *FreeRstr (Fluids)

 (prog (n lng list cfluids listfluids lab initload freeregs)
      (setq lab (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 2)) % two words per BndStk entry
      (setq freeregs '((reg t3)(reg t4)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids
      (setq n (wtimes2 4 (wdifference 2 lng)))
      (setq lng (wtimes2 lng 4)) % * addressingunitperitem
      (setq initload (list '(*move ($VIVfluid Bndstkptr) (reg t1))))

preload  (setq initload
       (progn (setq listfluids
               (if freeregs
                (nconc listfluids (cons (car freeregs) nil))
                (nconc listfluids (cons nil nil))) )
        (nconc initload
         (if freeregs
%%%      `((*move (indexed (reg t1) ,(wtimes2 4 n)) ,(car freeregs))) nil)
     `((*move (indexed (reg t1) ,n) ,(car freeregs))) nil)
       )   ))
       (setq n (wplus2 n 8))
       (when freeregs (setq freeregs (cdr freeregs)))
       (setq cfluids (cdr cfluids))

       (cond (cfluids (go preload)))

       (setq freeregs listfluids ) %end of preloadcode

       % freeregs contains the list of preloaded regs
       % and nil if not enough regs available

      (setq list `((sub   (Reg t1) ,lng (reg t2))
               (*move ($VIVfluid BndstkLowerBound) (Reg t5))
               (subcc (reg t2)(reg t5) (reg g0))
               (bl    (entry Bstackunderflow))
               (*move (Reg t2) ($VIVfluid BndstkPtr))  )) %start of code

     (setq list (append initload list))
     (setq n 0)

 loop
      (setq cfluids (car Fluids))
      (setq n (wplus2 n 8))

  % insert reloaded register or memory reference

      (setq list (append list
           (if (car freeregs) `((*move ,(car freeregs) ,cfluids ))
%                    `((*move (indexed (Reg t2) ,(wtimes2 4 n)) ,cfluids )))
                `((*move (indexed (Reg t2) ,n) ,cfluids )))

      )          )
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *freerstr)

(setq *unsafebinder t)   % has to save Registers across calls

(global '(*writingasmfile))

(de !*xjumpon (register lowerbound upperbound labellist)
    (for (from x lowerbound upperbound)
         (join
          `((subcc ,register ,x (reg g0))
            (be ,(pop labellist))
            (nop)))))

(de !*jumpon (register lowerbound upperbound labellist)
   (if *writingasmfile
       (*xjumpon register lowerbound upperbound labellist)
     (PROG (X LL LL2)
       (setq ll  (gensym))
       (setq ll2 (stringgensym))
       (SETQ X
        (if (and (weq lowerbound 0) (weq upperbound 31) *syslisp)
                  % jumpon on tags (most probably)
      `((sll ,register 3 (reg t5))  % 8 bytes per jumptable entry
        (move-relocatable-object ,ll2 (reg t1))
        (jmpl (displacement (reg t1) (reg t5)) (reg 0))
        (nop)
       ,ll2)
      `((subcc ,register ,upperbound (reg 0))
        (be!,a ,(lastcar labellist))
        (nop)
        (bpos!,a (label ,ll))
        (nop)
        (subcc ,register ,Lowerbound (reg t5))
        (bneg!,a (label ,ll))
        (nop)
        (sll (reg t5) 3 (reg t5))  % 8 bytes per jumptable entry
    (move-relocatable-object ,ll2 (reg t1))
        (jmpl (displacement (reg t1) (reg t5)) (reg 0))
        (nop)
       ,ll2) ) )
      Loop  (Setq x (nconc X `((BA ,(car Labellist)) ,(list 'nop))) )
            (setq Labellist (cdr Labellist))
            (cond (Labellist (go loop)))

            (setq x (nconc x `((*lbl (Label ,ll))) ))
            (return x)
)  )  )

(defcmacro !*jumpon)


% Call foreign fns (loaded by oload) by putting SymFunc cell addr in a reg,
% then doing "calls" instr indirect via the target address of the jmp
% instruction stored there.  Use old direct link for asm (kernel) compile.

(de *foreignlink (functionname functiontype numberofarguments)

% Following Sparc Architecture Manual, the Parameter past the six have
% to be passed in stack. We therefore need some extra space, we get it
% by doing an extra alloc.

 (if (wgreaterp numberofarguments 6)
     `((*alloc 30)            % This is a rare case (hopefully)
       (*move (reg 15) (reg o5)) (st (reg o5) (indexed (reg sp) 124))
       (*move (reg 14) (reg o5)) (st (reg o5) (indexed (reg sp) 120))
       (*move (reg 13) (reg o5)) (st (reg o5) (indexed (reg sp) 116))
       (*move (reg 12) (reg o5)) (st (reg o5) (indexed (reg sp) 112))
       (*move (reg 11) (reg o5)) (st (reg o5) (indexed (reg sp) 108))
       (*move (reg 10) (reg o5)) (st (reg o5) (indexed (reg sp) 104))
       (*move (reg 9)  (reg o5)) (st (reg o5) (indexed (reg sp) 100))
       (*move (reg 8)  (reg o5)) (st (reg o5) (indexed (reg sp) 96))
       (*move (reg 7)  (reg o5)) (st (reg o5) (indexed (reg sp) 92))
       (*move (reg 6) (reg o5))
       (*move (reg 5) (reg o4))
       (*move (reg 4) (reg o3))
       (*move (reg 3) (reg o2))
       (*move (reg 2) (reg o1))
       (*move (reg 1) (reg o0))
       (call (foreignentry ,functionname) ,numberofarguments)
       (nop)
       (*move (reg o0) (reg 1))
       (*dealloc 30))

   (let ((sequence nil))
     (setq sequence
               (cond ((equal numberofarguments 0) nil)
                     (t (pnth '((*move (reg 6) (reg o5))
                             (*movex argfive  (reg o4))
                                (*movex argfour  (reg o3))
                                (*movex argthree (reg o2))
                                (*movex argtwo   (reg o1))
                                (*movex argone   (reg o0)))
                              (difference 7  numberofarguments)))))
     (if (null (getenv "linux"))
      (setq sequence
       (append sequence
          (list (list 'call (list 'foreignentry functionname)
                                     numberofarguments)
                         (list 'nop))))
      (setq sequence
       (append sequence %  we have to restore g7 across foreigncall
          (list (list 'call (list 'foreignentry functionname)
                                     numberofarguments)
                     '(*move (reg g7) (reg l3))
                     '(*move (reg l3) (reg g7)))))
     )
     (setq sequence
       (append sequence
                  '((mov (reg o0) (reg 1))   %store return value in r1
                   ))))))

(defcmacro !*foreignlink)

(defcmacro *movex)
   % Movex is of type cmacro

(de *movex (src dst)
     (let (tmp)
      (cond ((sparc-register-p dst)
          (cond ((sparc-register-p src)  `((mov ,src ,dst)))
		((eqcar src 'sethiidloc)
			(if *writingfaslfile
				`((sethi-dummy ,(cadr src)))
			  (if  *writingasmfile
				`((sethi (hi (lshift (idloc ,(cadr src)) 10))
					,dst))
			`((sethi (immediate (idloc ,(cadr src))) ,dst)))))
                ((and (sparc-fluid-p src) (equal dst '(reg fsr))
                          *writingasmfile)
                         `((sethi (hi ,src) (reg t5))
                           (ld (displacement (reg t5) (lo ,src)) ,dst)))
                ((eqcar src 'quote)
                     (if *writingasmfile           %% asmout and lap generate
                         `((sethi (hi ,src) ,dst)  %% different code.
                           (add  ,dst (lo ,src) ,dst))
                         `((move-relocatable-object ,src ,dst))))
                ((or (sparc-absolute-p src) (sparc-fluid-p src))
                 (if *writingasmfile           %% asmout and lap generate
                     `((sethi (hi ,src) ,dst)  %% different code.
                       (ld (displacement ,dst (lo ,src)) ,dst))
                     `((move-relocatable-object-and-load ,src ,dst))))
                ((sparc-short-immediate-p src) `((mov ,src ,dst)))
                ((or (sparc-relocatable-immediate-p src)
                     (sparc-foreignentry-p src))
                 (if *writingasmfile           %% asmout and lap generate
                     `((sethi (hi ,src) ,dst)  %% different code.
                       (add ,dst (lo ,src) ,dst))
                     `((move-relocatable-object ,src ,dst))))
                ((sparc-immediate-p src)
                     `((sethi (hi ,src) ,dst)
                       (add ,dst (lo ,src) ,dst)))
                ((sparc-huge-displacement-p src)
                     `((sethi (hi ,(caddr src)) ,dst)
                       (add ,dst (lo ,(caddr src)) ,dst)
                       (ld (,(cadr src) ,dst) ,dst)))
                ((and (sparc-deferred-p src) (sparc-indirect-p (cadr src)))
                       (car (load-src-arg src dst)))
                ((sparc-indirect-p src) `((ld ,src ,dst)))
                (t  (print-error-status)
                    (error 0
                 (list "Unknown addressing mode for *movex:" src dst)))))
         ((sparc-register-p src)
          (cond ((or (sparc-absolute-p dst)
                     (sparc-fluid-p dst)
                     (sparc-foreignentry-p src))
                 (if *writingasmfile
                     `((sethi (hi ,dst) (reg t6))
                       (st ,src (displacement (reg t6) (lo ,dst))))
                     `((move-relocatable-object-and-store
                                     ,dst (reg t6) ,src))))
                ((and (sparc-deferred-p dst) (sparc-indirect-p (cadr dst)))
                 (setq tmp (load-src-arg (cadr dst) '(reg t6)))
                 (append (car tmp) `((st ,src (deferred (reg t6))))))
                ((sparc-huge-displacement-p dst)
                 `((sethi (hi ,(caddr dst)) (reg t6))
                   (add (reg t6) (lo ,(caddr dst)) (reg t6))
                   (st ,src (,(cadr dst) (reg t6)))))
                ((sparc-indirect-p dst) `((st ,src ,dst)))
                (t   (print-error-status)
                     (error 0
                        (list "Unknown addressing mode:" src dst)))))
         (t (print (list " Unsaeglich: Move " src dst))
))))
