%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  File: SPARC-LAPOPT-2
%
%  Description: Code Optimization
%               Reformation of jump sequences
%
%  Author:      H. Melenk, ZIB Berlin
%  Date:        22 Dec 1986
%
% Copyright (c) 1987,  Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% operation:
%
% remove jumps to jumps
% remove nop from delayed slot

(compiletime (progn
	(on fast-integers)
	(flag '(operands LapOptRemoveToteHosen lookForLabelForwards
	 lookForLabelBackwards label-from  read-label) 'internalfunction)))
 
(fluid '(*ConditionalJumps* *unconditionalJumps* *goodjumps* % *badjumps*
         *nonMovables* *alljumps* *lapopt))
 
%(setq *badjumps*    % 
%     '((BNE . BNE!,a) (BE . BE!,a) (BG . BG!,a)
%       (BLE . BLE!,a) (BGE . BGE!,a) (BL . BL!,a) (BGU . BGU!,a)
%       (BLEU . BLEU!,a) (BCC . BCC!,a) (BCS . BCS!,a) (BPOS . BPOS!,a) 
%       (BNEG . BNEG!,a) (BVC . BVC!,a) (BVS . BVS!,a)
%  ))
 
(setq *goodjumps* '( BA BNE!,a BE!,a BG!,a BLE!,a BGE!,a BL!,a BGU!,a
         BLEU!,a BCC!,a BCS!,a BPOS!,a BNEG!,a BVC!,a BVS!,a))

(setq *conditionalJumps*
      '((BE!,a . BNE!,a) (BNE!,a . BE!,a) (BLE!,a . BG!,a) (BG!,a . BLE!,a)
        (BL!,a . BGE!,a) (BGE!,a . BL!,a) (BLEU!,a . BGU!,a) (BGU!,a . BLEU!,a)
        (BCS!,a . BCC!,a) (BCC!,a . BCS!,a) (BNEG!,a . BPOS!,a)
        (BPOS!,a . BNEG!,a) (BVS!,a . BVC!,a) (BVC!,a . BVS!,a)
   ))
 
(setq *unconditionalJumps* '(BA JMPL CALL))

(setq *nonMovables* (append *goodjumps!* '(
 MOVE-RELOCATABLE-OBJECT MOVE-RELOCATABLE-OBJECT-and-load
 MOVE-RELOCATABLE-OBJECT-and-store nop *entry call jmpl save restore)))

(setq *alljumps* (append *goodjumps!* *unconditionalJumps*))

(compiletime (ds NOPp (x) ((lambda(h) (and (pairp h)(eq (car h) 'NOP))) x)))

(de LapOptReformJumps (code)
  (let (rcode rrcode inst inst2 x y z cond jump inv lab lastinst delayslot)
       (when (NOT *LAPOPT) (return code)) % Main switch

%        % first phase% change (any (without reg))
%                              (ld (reg)
%                              (any (using reg))
% into (ld ..) (any ..) (any ..)

      (while code
         (setq inst (pop code))
         (when (and (pairp inst) rcode (pairp (setq lastinst (car rcode)))
                (eqcar lastinst 'ld) (member (caddr lastinst) (operands inst))
                (cdr rcode) (pairp (setq inst2 (cadr rcode)))
                (not (member (caddr lastinst) (operands inst2)))
                (or (atom (cddr rcode)) (atom (caddr rcode))
                    (not (member (caaddr rcode) *alljumps*)))
                % dont "steal" from delay slot
                (or (atom (cdr inst2)) 
                    (not (member (lastcar inst2) (operands lastinst))))
                % target of first instr must not be needed for second
                (not (and (eqcar inst2 'st)
                          (equal (caddr inst2) (cadr lastinst))))
                % no intermix of ld and st of same object
                (not (member (car inst2) *nonMovables*)))
                (setq x (pop rcode)) (setq y (pop rcode))
%	(print (list "tausche" inst x y))
                (push x rcode) (push y rcode))
        (push inst rcode))

      (setq code (reversip rcode)) (setq rcode nil)

        % second phase: look for jumps to jumps and for negation jumps
     (setq delayslot 0)
     (while code
          (setq inst (pop code)) 
          (when (and (pairp inst)    % negation jump?
                (setq x (atsoc (car inst) *conditionalJumps*)) 
                (setq lab (label-from  inst))
                 code (NOPp (car code) )
                (pairp (setq y (cdr code))) 
                (pairp (setq inst2 (car y))) (eq (car inst2) 'BA)
                (cdr y) (NOPp  (cadr y))
                (setq z(cddr y)) (equal (car z) lab))
            (setq inst (cons(cdr x) (cdr inst2)))
            (setq code (cons '(NOP) z))     )
          (while(and (pairp inst)    % jump to jump ?
                (memq (car inst) *goodJumps*)
                (setq lab (label-from inst))  
                code (equal (car code) '(NOP))
                (setq x (or (lookForLabelForwards code lab)
                            (lookForLabelBackwards rcode lab)))
                (eqcar (car x) 'BA)
                (NOPp (cadr x) )
             (setq inst (cons (car inst) (cdar x)))    )  )
          (push inst rcode))  
     (setq code (reversip rcode)) (setq rcode (setq rrcode (cons nil nil))) 
          % third phase: fill in delayed instructions
     (while code
         (setq inst (pop code)) 
         (setq delayslot (cond ((eq delayslot 3) 2)
                               ((eq delayslot 2) 1)
                               ((eq delayslot 1) 0)
                               (t 0)))
	 (while(and (pairp inst)    % jump to jump ?
                (memq (car inst) *goodJumps*)
                (setq lab (label-from inst))  
                code (NOPp  (car code) ) 
                (setq x (or (lookForLabelForwards code lab)
                            (lookForLabelForwards rcode lab)))
                (setq inst2 (car x))
                (pairp inst2)  
                (not (memq (car inst2) *nonMovables*)))
                                     % we insert a label
            (when(not(atom(setq lab (cadr x))))
                 (setq lab (gensym)) (rplacd x (cons lab (cdr x))))
            (rplacd rrcode  (setq z `((,(car inst) ,lab))))
     % used to be 3
	    (setq delayslot 2) %(setq lastinst (car rrcode))
            (setq rrcode z)
            (setq inst (car x))
            (setq code (cdr code)))
        (when % last case
              % (something) (BR lbl)(NOP)
              % -> (BR lbl) (something)
              % if (something) is no label and is movable
            (and (pairp inst) (memq (car inst) *unconditionalJumps*)
                 code 
                 (NOPp (car code))
                      % if not 2nd successor of a jump
                 %or (null lastinst)
                 %   (and (pairp lastinst)
                 %        (not (memq (car lastinst) *allJumps*))))
                 (eq 0 delayslot)
                 (pairp (car rrcode)) 
		 (null (intersection (operands inst) (operands (car rrcode))))
			% for (jmpl (displacement (reg t4) 0) (reg 15))
			% new calling sequence and apply s
		 (not(memq (caar rrcode) *nonMovables*)))
            (setq code (cdr code))
            (setq delayslot 3) % (setq lastinst inst)
            (setq y (car rrcode)) (rplaca rrcode inst)(setq inst y)
            )
        (rplacd rrcode  (setq z (cons inst nil)))
        (when (and (pairp inst) (memq (car inst) *alljumps*))
               (setq delayslot 3)) % this instruction has a delay slot
        (setq rrcode z))
           % done
        (LapOptRemoveToteHosen (cdr rcode))
  ))

% since the jumps are (BA hugo) and not (ba (label hugo)
% and it does not handle the jumps backward correctly (see tak example
(de LapOptRemoveToteHosen (code)
   (prog (l labels rcode inst prevjump1 prevjump2 alllabs)
       (setq alllabs (mapcan code
                   (function (lambda(inst) (when (atom inst) (list inst))))))
       (mapc code (function (lambda(inst)
            (when (pairp inst)
             (setq l (read-label inst alllabs))
             (when (not(member l labels)) (setq labels (cons l labels)))))))
       (while code
              (setq inst (pop code))
              (while (and (atom inst)                   % is a label
                          (not (member inst labels))    % no reference
                          prevjump2)                    % previous is a jump
                     (while (and code (pairp (setq inst (pop code))))))
              (setq prevjump2 prevjump1)
              (setq prevjump1 (when (pairp inst)
                                    (or (eqcar inst 'BA) 
                                        (and (eqcar inst 'JMPL) 
                                             (neq (cadr (caddr inst)) 15))))) 
%                                   (memq (car inst) '(BA JMPL))))
              (push inst rcode))
       (return(reversip rcode)) ))

(de lookForLabelForwards(code lab)
       (cond ((null code) nil)
             ((equal (car code) lab)(cdr code))
             (T (lookForLabelForwards (cdr code) lab))))
 
(de lookForLabelBackwards(code lab)
       (cond ((null code) nil) 
             ((null (cdr code)) nil)
             ((null (cddr code)) nil)
             ((equal (caddr code) lab)(list (cadr code) (car code)))
             (T (lookForLabelBackwards (cdr code) lab))))
         
(de label-from (inst) 
     (setq inst (cadr inst))
     (if (atom inst) inst (cadr inst)))
     
(de read-label (u alllabs)
    (cond ((atom u) (if (member u alllabs) u nil))
          ((read-label (car u) alllabs))
          (T (read-label (cdr u) alllabs))))

(de operands (x)         %simplified form , only looking for regs
   % extract the operands from x
   (cond ((atom x) nil)
         ((member x '((reg zero))) nil)
         ((or (memq (car x) '(reg))) (list x))
         ((memq (car x) '(scaled displacement label wconst indexed indirect))
          (operands (cdr x)))
         (t (nconc (operands (car x)) (operands (cdr x))))))

