%      Lap Code optimizers

(compiletime (load if-system))
(compiletime (progn (load useful) (on fast-integers)))

(fluid '(!*lapopt))

(de lapopt (u)
   % lapopt for cmacros
    (prog()
        (when (not *lapopt) (return u))
        (when (or (eq *lapopt t)
                  (and (numberp *lapopt)
                       (not (izerop (wand *lapopt 1)))))
              (setq u (lapoptFrame2a6 u)))
        (when (or (eq *lapopt t) 
                  (and (numberp *lapopt) 
                       (not (izerop (wand *lapopt 2))))) 
              (setq u (lapoptcarcdr u)))
         (return u)))
 
(de lapopt2 (u)
   % lapopt for instructions
    (prog()
        (when (not *lapopt) (return u))
     (when (or (eq *lapopt t)
                  (and (numberp *lapopt)
                       (not (izerop (wand *lapopt 4)))))
              (setq u (lapoptreformloads u)))
        (return u)))
     
%%----------------------------------------------------------------- 
% 
%   map frame with highest usage to register A6
%   which is saved automatically on the CONVEX and not used by LISP

(de lapoptFrame2a6(u)
    (let ((f (lapoptFrame2a6max u)))
         (if f 
             (lapoptFrame2a6subst u f) 
             u)))
 
(de lapoptFrame2a6max(u)
     (car (lapoptFrame2a6maxsel (lapoptFrame2a6count u nil) '(nil . -1))))
 
(de lapoptFrame2a6count(u r)
   (cond ((atom u) r)
         ((and (eqcar u 'frame) (cdr u) (numberp (cadr u)))
           (let ((a (assoc u r)))
                (if a 
                    (progn (rplacd a (add1 (cdr a))) r)
                    (cons (cons u 1) r))))
         (t (lapoptFrame2a6count (cdr u)
                (lapoptFrame2a6count (car u) r)))))
 
(de lapoptFrame2a6maxsel(l p)
     (if (null l) 
         p
         (lapoptFrame2a6maxsel (cdr l)
                 (if (greaterp (cdar l) (cdr p)) (car l) p))))

(de lapoptFrame2a6subst(u f)
   (let (x n)   
       (setq u (subst '(reg a6) f u))
       (setq n (cadr f))
       (when (setq x (lapoptFrame2a6finpush u n))
             (rplacd x (cons (car x) (cdr x)))
             (rplaca x `(*move ,(cadr (car x))(reg a6))))
       u))
 
(de lapoptFrame2a6finpush(u n) 
    (cond ((or (wgreaterp n 2) (null u)) nil) 
          ((eqcar (car u) '*push)
           (if (eq n 1) u (lapoptFrame2a6finpush (cdr u) 1))) 
          (t (lapoptFrame2a6finpush (cdr u) n) )))
 

%%-----------------------------------------------------------------
%
%   Car/Cdr access optimization

(de lpmatch(f p) (lpmatch1 f p '( nil . nil)))

(de lpmatch1(f p a)
   % f: form
   % p: pattern
   % a: binding (alist)
 (cond ((null a) nil)
       ((eq p 'any) a)
       ((atom p)(cond ((equal f p) a)(t nil)))
       ((atom f) nil)
       ((member (car p) '(frame reg))
        (cond ((and (eq (car f)(car p))
                    (numberp (cadr f)))
               (cons (cons (cadr p) (cadr f)) a))
              (t nil)))
       (t (lpmatch1 (cdr f)
                    (cdr p)
                    (lpmatch1 (car f)(car p) a)
))     )  )
 
(de lpval(id al) (cdr (atsoc id al)))
 
% the swtich *pushafter defines, if additional push instructions
% have to be inserted behind or in front of the original push sequence
%        behind if frames count from bottom of frame (CONVEX),
%        not behind if frames count from actual stack position (68020)

(fluid '(!*pushafter))
 
(if_system CONVEX
   (setq !*pushafter t)
   (setq !*pushafter nil)
)

(de lapoptcarcdr(u)
   (prog(alloc nalloc i x y z r f code pos labels regframes done coll push)
      (setq code u)
      (while code
        (setq pos code)
        (setq i (pop code))

              % reasoning on allocsize
        (when (null alloc)
         (cond
          ((eqcar i '!*alloc)
           (setq alloc i)
           (setq nalloc (cadr alloc)))

          ((eqcar i '!*push)
           (setq nalloc 1)
           (when (regp (cadr i)) (push (cons (cadr (cadr i)) 1) regframes))
           (setq x (car code))  % next instruction
           (when (eqcar x '!*push)
                (setq nalloc 2) 
                (when (regp (cadr x)) (push (cons (cadr (cadr x)) 2) regframes))
                (when !*pushafter (setq i (pop code)))
           )
           (setq alloc (list nil nalloc))
           (setq push (if !*pushafter code pos)) 
          )))

             % collect labels of far jumps
       (when (eqcar (cadr i) 'label)
             (push (cadr (cadr i)) labels))

             % collect instructions which contain allocsize
       (when (and (pairp i) (memq (car i) '(!*linke !*exit))) (push i coll))

             % look for (*move (car (frame f)) (reg r))
       (when 
           (setq f (carcdrframe i regframes))
                % look for more references in the subsequent code
           (when (setq z (citescarcdr f code labels))
             (setq done t)
             (when (cdr z)
                % cdr needed: setup macro *carcdr and save cdr
                (prin2 " # ")
                (setq y
                %  (if (equal (cdr z) `((!*move (cdr (frame ,f))(frame ,f))))
                %      `(frame ,f)
                       (auxframe alloc)
                %  )
                )
                (rplaca pos
                   `(!*carcdr ,(cadr(cadr i)) ,(caddr i) ,y))
                (substip y `(cdr (frame ,f)) (cdr z))
             )
             (when (car z)
                % car needed more than once; save the value
                (prin2 " @ ")
                (setq y (auxframe alloc))
                (rplacd pos
                  (cons
                    `(!*move ,(caddr i) ,y)
                     code))
                (substip y `(car (frame ,f)) (car z))
        )))
 
        (setq regframes (updateregisters i regframes))

      ) %while

      (when done 
            (updateallocsize (cadr alloc) coll)
            (when push (insertpush push (difference (cadr alloc) nalloc)))
            (when *plap (terpri) (terpri) (mapc u 'print))
            )

      (return u)))

(de carcdrframe(i regframes)
    % inspect instruction i, if ti is a loads a car of a register
    % if yes, return the number of the frame
     (let (x y z)
         (or
            (and
               (setq x(lpmatch i '(!*move (car (frame f))(reg r)) ))
               (lpval 'f x)
            )
            (and
               (setq x(lpmatch i '(!*move (car (reg s)) (reg r)) ))
               (setq z (lpval 's x))
               (setq y (assoc z regframes))
               (cdr y)
           ))))

(de updateregisters (i regframes) 
   %  maintenance of the register-frame relations
  (prog (f r x y)            
 
     (when (setq r (destroysreg i))
           (when (setq y (assoc r regframes)) (setq regframes (delete y regframes))))
 
     (when (setq f (destroysframe i))
           (when (setq y (rassoc f regframes)) (setq regframes (delete y regframes))))
 
     (when (setq x (lpmatch i '(!*move (reg r)(frame f))))
           (setq r (lpval 'r x)) 
           (setq f (lpval 'f x))
           (push (cons (lpval 'r x)(lpval 'f x)) regframes))

                % all registers get lost
     (when (or (atom i) (memq (car i)'(!*link !*linke !*progbind !*freerstr)))
           (setq regframes nil))
 
     (return regframes)))

(de updateallocsize (n c)
   % replace original allocsize in *exit and *linke instructions
    (cond ((null c) nil)
          (t (updateallocsize n (cdr c))
             (setq c (car c))
             (rplaca (cdr c) n) )))

(de insertpush(p n)
   % add some instructions (*push nil) in order to increase
   % allocsize
    (cond ((zerop n) nil)
          (t (rplacd p (cons (car p) (cdr p)))
             (rplaca p '(!*push (quote nil)))
             (insertpush p (sub1 n)) )))

(de substip(x u s)
  % subst in place
  (cond ((atom s) nil)
        ((equal (car s) u)(rplaca s x)
                          (substip x u (cdr s)) )
        (t (substip x u (car s)) (substip x u (cdr s)) )))

(de auxframe(alloc)
    (prog (n)
      (setq n (add1 (eval (cadr alloc))))
      (rplaca (cdr alloc) n)
      (return `(frame ,n))))


(de citescarcdr(f code nastylabels)
  (prog (cars cdrs i)
    (while code
     (setq i (pop code))
     (cond
       ((or (atom i)
            (and (eqcar i '*lbl)
                 (setq i (cadr (cadr i)))))
        (when 
           (or (member i nastylabels)
               (badbwtarget i f code nastylabels)  
           )
           (setq code nil)
       ))
       ((citescarcdr1 'car f (cdr i))(push i cars))
       ((citescarcdr1 'cdr f (cdr i))(push i cdrs))
     )
     (cond
       ((destroysframe f i) (setq code nil))
    )) 
   (return (if (or cars cdrs) (cons cars cdrs) nil))
))
 
(de citescarcdr1(op f i)
  (cond ((atom i) nil)
        ((and (eqcar i op)
              (eqcar (cadr i) 'frame)
              (equal (cadr(cadr i)) f)
         ) t)
        (t (or (citescarcdr1 op f (car i))
               (citescarcdr1 op f (cdr i))
) )     )  )

(de badbwtarget(lab f code nastylabels)
  % a label is bad, if there is a jump backwards to it from a
  % point, where the frame f might have changed.

  (prog(fail i destroyed)
    (while (and code (not fail))
      (setq i (pop code))
      (cond

            % a jump-into label: frame no longer reliable
         ((atom i)
          (when (or (member i nastylabels)
                    (badbwtarget i f code nastylabels)
                )
                (setq destroyed t) ))
     
            % frame overwritten explicitly
         ((equal (destroysframe f) i) (setq destroyed t))

            % a jump to this label is prohibitive
         ((and (cdr i) 
               (eqcar (cadr i) 'label)
               (eq (cadr(cadr i)) lab))
          (setq fail destroyed))
 
            % an exit from the sequence reopens the game
         ((or (memq (car i) '(!*exit !*linke))
              (and (eqcar i '!*jump)
                   (member (cadr (cadr i)) nastylabels)
          )   )
          (setq destroyed nil))
    ) )
    (return fail)
) )     
             
(de destroysframe(i) (destroysplace 'frame i))
 
(de destroysplace(place i)
  % predicate: instructiion i overwrites a frame or reg 
 (and (not(atom i))
   (or (and (eqcar i '!*move)
            (setq i (caddr i))
            (eqcar i place)
            (cadr i))
       (and (not (eqcar i '!*move))
            (setq i (cadr i))
            (eqcar i place)
            (cadr i) ) )))

(de destroysreg(i) (destroysplace 'reg i))
            
(defcmacro *carcdr)

(de *carcdr(a b c)
  % we know that a is a register or a frame, b is a register and c
  % is a frame. No resolve will be necessary
    `((*move ,a (reg a1))
      (ld!.l (indexed (reg a1) 0) ,b)
      (*move ,b ,c)
      (*wshift ,b (wconst 32)) ))
 
%----------------------------------------------------------------------
 
(fluid '(jumps*))
 
(setq jumps* '(br  bra!.t bra!.f  brs!.t brs!.f
               jmp jmps!.t jmps!.f  jmps!.t jmps!.f
               calls calls! @ *entry))


(de LapOptReformLoads (code)
  (let (rcode rrcode inst inst2 x y z cond jump inv lab lastinst ol)
      (while code
	 (setq inst (pop code))
	 (when (and (pairp inst) 
                    rcode 
                    (pairp (setq lastinst (car rcode)))
		    (eqcar lastinst 'ld!.w) 
		    (member (caddr lastinst) (operands inst))
		    (cdr rcode) 
		    (pairp (setq inst2 (cadr rcode)))
		    (not (member (caddr lastinst) (operands inst2)))
                    (not (memq (car inst2) jumps*))
                    (setq ol (operands lastinst))
		    (or (atom (cdr inst2))
                        (if(eq (car inst2) 'PSH!.W)
                           (and
                               (not (member '(INDEXED (REG A1) 0) ol))
                               (not (member '(INDEXED (REG A1) 4) ol))))
		           (not (member (caddr inst2) ol)))
		% target of first instr must not be needed for second
		    (not (and (eqcar inst2 'st!.w)
		              (equal (caddr inst2) (cadr lastinst))))
                )
		(setq x (pop rcode)) (setq y (pop rcode))
		(push x rcode) (push y rcode))
	(push inst rcode))

      (setq code (reversip rcode)) 
      (return code)
  ))


(de operands (x)	 %simplified form , only looking for regs
   % extract the operands from x
   (cond ((atom x) nil)
         ((member x '((reg zero))) nil)
	 ((or (memq (car x) '(reg))) (list x))
         ((memq (car x) '(scaled displacement label wconst indexed indirect))
          (operands (cdr x)))
         (t (nconc (operands (car x)) (operands (cdr x))))))


