%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PC:FASLOUT.SL
% SCCS:         %A% %G% %U%
% Title:        Top level of fasl file writer
% Author:       Eric Benson
% Created:      16 February 1982
% Modified:     23 Jan 1985 0937-PST (Brian Beach)
% Status:       Experimental
% Mode:         Lisp
% Compiletime:  PL:FAST-VECTOR.B PL:FASL-DECLS.B
% Runtime:
%
% (c) Copyright 1987, Konrad Zuse Zentrum Berlin, for modifications     .
% (c) Copyright 1983, 1984, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1982, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 18 Aug 88 W. Neun, ZIB Berlin
%  Disabled lapopt for initcode, since not needed there.
% 18 Nov 86 W. Neun, ZIB Berlin
%  Modifications to updatebittable and faslout to dynamically enlarge
%  fasl-space in case of lack
% 3 March 1986 (Herbert Melenk, ZIB Berlin)
%  Modifications for CRAY: FASL addresses are calculated in
%  parcels.
%  BinaryWrite.. routines are replaced by FaslWrite... in order to
%  allow production of fasl files as library members
%
% 23 Jan 1985 0937-PST (Brian Beach)
%  Changed FASLOUT so that it returns T again.
% 15-Jan-85 15:49:52 (Brian Beach)
%  Added a fluid variable called FASL-IDNUMBER-PROPERTY* which is the name
%  of the property that id numbers are stored under.  This is to fix a problem
%  that occurrs when a compile is abnormally terminated, and another one
%  started where id numbers in the new compile are confused with id numbers
%  in the terminated compile that weren't cleaned up.
% 14-Nov-84 14:20:49 (Brian Beach)
%  Changed FASLABORT accoring to Utah fix so that the output file is closed
%  after the flags are reset, so that the system is left in a reasonable state
%  even if the file cannot be closed.
% 24-Oct-84 15:30:01 (Alan Snyder)
%  Install changes from Bob Morain:
%    Now turns off internalfunction flag at the end of the file to restore
%    compiler to good state.  Added a new internal function
%    DEALLOCATEFASLSPACES.  Flagged FASLABORT 'ignore to make it more useful;
%    made it call DEALLOCATEFASLSPACES.  Flush flagging of STARTUPTIME, which
%    is obsolete.  Changed FASLOUT-MAGIC to (COMPILER-CONSTANT
%    'FASLOUT-MAGIC).  Added FASL-PREEVAL* to parameterize the property looked
%    up by DFPRINTFASL; DFPRINTFASL can now be used (and is used) by ASMOUT.
%    Made FASL-FUNCTION call COMPD, which seems to do pretty much the same
%    things that FASL-FUNCTION did.  Remove redundant binding of
%    *WRITINGFASLFILE by DFPRINTFASL.
%  In addition:
%    Make FASLEND call FASLABORT to reduce duplicated code.
% 19-Oct-84 08:37:49 (Alan Snyder)
%  Fix another bug in 02-Oct-84 change: LOADTIME was calling FASL-FORM, which
%  interprets the IGNORE and EVAL flags (which implement COMPILETIME and
%  BOTHTIMES).  It appears that the behavior desired by PSL programmers is for
%  LOADTIME to force compiler output and prevent compiler execution, as in
%  (LOADTIME (ON FOO)), where ON is implicitly COMPILETIME.  This behavior
%  differs from Common Lisp EVAL-WHEN, where an (EVAL-WHEN (LOAD) ...) is a
%  no-op to the compiler (the body is compiled normally).  My fix is to add a
%  parameter to fasl-from and the faslpreeval functions and and pass this down
%  (to handle macro calls).
% 15-Oct-84 11:15:49 (Alan Snyder)
%  Fix bug in 02-Oct-84 change: Add ERRSET around invocation of TOPLOOPEVAL*.
% 02-Oct-84 09:23:07 (Alan Snyder)
%   Restructured DFPRINTFASL for much improved clarity.  Flushed the redundant
%   checking for COMPILE flag on function type (EXPR, etc.). Changed to use
%   the existing PREFASLEVAL hook for special treatment of PUTD, DE, etc.,
%   instead of special case code.  Added the EVAL and IGNORE flag handling in
%   DFPRINTFASL (instead of DSKIN and TOPLOOP), for proper interaction with
%   macro expansion; macros can now expand into COMPILETIME, etc.  (Note
%   30-Apr-82 change, for amusement.)  Changed to use TOPLOOPEVAL* instead of
%   EVAL, for multiple dialect support.  Added warning message for non-list
%   form given to DFPRINTFASL.  Removed PACKAGE attribute from the file
%   header.
% 04-Jun-84 13:46:13 (Brian Beach)
%  Replaced compiletime dskin of SYSTEM-FASLOUT with compiletime load of
%  FASL-DECLS.  New constant model.  Replaced use of 2048 with constant
%  first-local-id-number.
% 12 Apr 1984 1411-PST (Nancy Kendzierski)
%  Changed if_system unix to if_system vax and relative path load to shell
%  variable $pv.  Added hpux200 compiletime load.
% 28 Mar 1984 1702-PST (Nancy Kendzierski)
%  Changed references to addressingunitsperitem from const to wconst.
%  Changed references to faslout-magic (previously FASL_MAGIC_NUMBER) from
%  const to wconst.
%  Moved (load monsym) to p20:system-faslout.sl where it belonged.
% 11-Jan-84 17:19:45 (Cris Perdue)
%  Modified codefiletrailer to accept a real string with a
%  header tag in its header word.  FASL file will always contain a
%  count for each string with no random extra bits set in that word.
% 06-Dec-83 10:00 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime
 (progn (flag '(codefileheader codefiletrailer allocatefaslspaces
                         deallocatefaslspaces)
            'internalfunction)
      (define-constant first-local-id-number 2048)
      (define-constant BitTable-Entries-Per-Word  32)
      )) 
 
(fluid '(*writingfaslfile
       *lower
       *quiet_faslout
       dfprint*
       uncompiledexpressions*
       modulename*
       codeout*
       initoffset*
       currentoffset*
       faslblockend*
       maxfasloffset*
       bittableoffset*
       faslfilenameformat*
       fasl-idnumber-property*
       *LapOpt
       *LapOptOptions*))
 
(setq faslfilenameformat* "%w.b")
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DFPRINTFASL - the external entry point to compile each form
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(toploopeval* fasl-preeval*))
 
(de dfprintfasl (u)
  % Called by top-loop, dskin, dfprint to compile a single form.
  (fasl-form u NIL))
 
(de fasl-form (u loadtime?)
  % Compile a simple S-EXPRESSION, maybe EVAL it.
  % If LOADTIME? is true, then do not obey the IGNORE and EVAL flags.
 
  (if (not (pairp u))
    (when u (errorprintf "*** Warning: non-list form ignored: %p" u))
    % Otherwise:
    (let ((fn (car u)) fcn)
      % Evaluate the expression if FN is flagged IGNORE or EVAL.
      (when (and (not loadtime?)
             (idp fn)
             (or (flagp fn 'ignore) (flagp fn 'eval)))
      (errset (funcall toploopeval* u) t))
      (cond
       % Do not compile if FN is flagged IGNORE.
       ((and (not loadtime?) (idp fn) (flagp fn 'ignore)) NIL)
       % Check for calls of special flagged function:
       ((and (idp fn) (setq fcn (get fn fasl-preeval*)))
      (funcall fcn u loadtime?))
       % Check for macro call:
       ((and (idp fn) (setq fcn (getd fn)) (equal (car fcn) 'macro))
      (fasl-form (funcall (cdr fcn) u) loadtime?))
       (t
      (saveuncompiledexpression u)
      NIL
      )))))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Special cases for FASL:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(put 'de 'faslpreeval
     (function (lambda (u loadtime?) (fasl-define u 'expr))))
(put 'df 'faslpreeval
     (function (lambda (u loadtime?) (fasl-define u 'fexpr))))
(put 'dm 'faslpreeval
     (function (lambda (u loadtime?) (fasl-define u 'macro))))
(put 'dn 'faslpreeval
     (function (lambda (u loadtime?) (fasl-define u 'nexpr))))
(put 'putd 'faslpreeval
     (function (lambda (u loadtime?) (fasl-putd u))))
 
(de fasl-define (u ty)
  % Compile a DE/DF/DM/DN.
  (fasl-function (cadr u) (cons 'lambda (cddr u)) ty))
 
(de fasl-function (nam u ty)
  % Compile a named function definition.
  (when (equal ty 'macro) % Macros get defined now (interpretively)
    (let ((*comp NIL))
      (putd nam ty u)))
  (compd nam ty u)
  (fasl-writename nam)
  )
 
(de fasl-writename (nam)
  % Write function name after compiling it.
  (when (wgreaterp (posn) 0)
    (writechar (char blank)))
  (prin1 nam)
  )
 
(de fasl-putd (u)
  % Compile a PUTD
  (let ((nam (cadr u))
      (ty (caddr u))
      (fn (cadddr u))
      )
    (cond ((and (eqcar nam 'quote)
            (eqcar ty 'quote)
            (memq (cadr ty) '(expr fexpr macro nexpr))
            (or (eqcar fn 'function)
                (eqcar fn 'quote)))
         (fasl-function (cadr nam) (cadr fn) (cadr ty)))
        (t
         (saveuncompiledexpression u)
         ))))
 
% Seems the wrong place for this:
(flag '(deflist flag fluid global remflag remprop unfluid) 'eval)
 
(de faslpreevalloadtime (u loadtime?)
  (fasl-form (cadr u) T))
 
% remove LOADTIME
(put 'loadtime 'faslpreeval 'faslpreevalloadtime)
 
(put 'bothtimes 'faslpreeval 'faslpreevalloadtime)
 
% used in kernel
% A few things to save space when loading
(put 'flag 'faslpreeval
     (function (lambda (u loadtime?)
                       (if (eqcar (second u) 'quote)
                         (fasl-form
                          (cons 'progn
                           (foreach x in (second (second u)) collect
                            (list 'flag1 (mkquote x) (third u)))) loadtime?)
                         (saveuncompiledexpression u)))))
 
(put 'fluid 'faslpreeval
     (function (lambda (u loadtime?)
                       (if (eqcar (second u) 'quote)
                         (fasl-form
                          (cons 'progn
                           (foreach x in (second (second u)) collect
                            (list 'fluid1 (mkquote x)))) loadtime?)
                         (saveuncompiledexpression u)))))
 
(put 'global 'faslpreeval
     (function (lambda (u loadtime?)
                       (if (eqcar (second u) 'quote)
                         (fasl-form
                          (cons 'progn
                           (foreach x in (second (second u)) collect
                            (list 'global1 (mkquote x)))) loadtime?)
                         (saveuncompiledexpression u)))))
 
(put 'deflist 'faslpreeval
     (function (lambda (u loadtime?)
                       (if (eqcar (second u) 'quote)
                         (fasl-form
                          (cons 'progn
                           (foreach x in (second (second u)) collect
                            (list 'put (mkquote (first x)) (third u)
                             (mkquote (second x))))) loadtime?)
                         (saveuncompiledexpression u)))))
 
(put 'progn 'faslpreeval
     (function (lambda (u loadtime?)
                       (foreach x in (cdr u) do (fasl-form x loadtime?)))))
 
(put 'lap 'faslpreeval
     (function (lambda (u loadtime?)
                       (if (eqcar (cadr u) 'quote)
                         (lap (cadr (cadr u)))
                         (saveuncompiledexpression u)))))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FASLOUT and friends:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(setq uncompiledexpressions* (cons nil nil))
 
(de saveuncompiledexpression (u)
  (if (atom u)
    nil
    (tconc uncompiledexpressions* u))
  nil)
 
(de faslout (fil)
  (setq modulename* fil)
  (unless *quiet_faslout
    (if (not (funboundp 'begin1))
      (progn (prin2t "FASLOUT: IN files; or type in expressions")
             (prin2t "When all done execute FASLEND;"))
      (progn (prin2t "FASLOUT: (DSKIN files) or type in expressions")
             (prin2t "When all done execute (FASLEND)"))))
  (setq codeout*
        (binaryopenwrite (bldmsg faslfilenameformat* modulename*)))
  (codefileheader)
  (setq dfprint* 'dfprintfasl)
  (setq fasl-preeval* 'faslpreeval)
  (setq *writingfaslfile t)
  (setq *defn t)
  (setf fasl-idnumber-property* (intern (concat "FASL-IDNUMBER-" (stringgensym))))
  t
  )
 
(de faslend ()
  (if (not *writingfaslfile)
    (stderror "FASLEND not within FASLOUT")
    (prog  (*LapOpt *LapOptOptions*) % no lapopt for initcode
         (compileuncompiledexpressions)
         (codefiletrailer)
         (return (faslabort))
) ) )
 
(flag '(faslend) 'ignore)
 
% FaslAbort.  Abort the Fasl process cleanly.  The code file will be closed
% and the various flags will be reset.
 
(de faslabort ()
  (if (not *writingfaslfile)
    (stderror "FASLAbort not within FASLOUT")
    (progn (setq uncompiledexpressions* (cons nil nil))
         (deallocatefaslspaces)
           (setq *writingfaslfile nil)
         (setq fasl-preeval* nil)
           (setq dfprint* nil)
           (setq *defn nil)
           (binaryclose codeout*)
         )))
 
(flag '(faslabort) 'ignore)
 
(de compileuncompiledexpressions ()
  (errorprintf "*** Init code length is %w"
               (length (car uncompiledexpressions*)))
  (dfprintfasl (list 'de '**fasl**initcode** 'nil
                     (cons 'progn (car uncompiledexpressions*)))))
 
(de codefileheader ()
  (binaryWrite codeout* faslout-magic)
  (allocatefaslspaces))
 
(fluid '(codebase* bittablebase* orderedidlist* nextidnumber*))
 
 % for CRAY machines the fasl addresses have to be calculated
 % as parcel addresses. That is similar to those machines which
 % have a value addressinunitsperitem=4. Unfortunately
 % the compatible formulation with this constant cannot be use
 % for Cray beause of the different addressing modes parcel/word.
 %
 % bittablebase*, faslblockend*, codebase* are calculated as
 % parcels.
 % maxfasloffset* is calculated in words
 %
 % Definition of some macros for address conversion between
 % words and parcels:
 
(compiletime (progn
  (ds pa(u) (wshift (inf u) 2))    % means: (times u 4)
  (ds wo(u) (wshift (inf u) -2))   % means: (quotient u 4)
)  )
 
 
(de findidnumber (u)
  (prog (i)
        (return (cond ((ileq (setq i (idinf u)) 128) i)
                      ((setq i (get u fasl-idnumber-property*)) i)
                      (t (put u fasl-idnumber-property* (setq i nextidnumber*))
                         (setq orderedidlist* (tconc orderedidlist* u))
                         (setq nextidnumber* (iadd1 nextidnumber*)) i)))))
 
(de faslid2string (x)
  % Stub to permit redefinition for "limited" package system
  (id2string x))
 
(de codefiletrailer ()
  (prog (s)
        (systemfaslfixup)
        (binarywrite codeout* (idifference (isub1 nextidnumber*)
                                 first-local-id-number))
        % Number of local IDs
        (foreach x in (car orderedidlist*) do
                 (progn (remprop x fasl-idnumber-property*)
                        (setq x (strinf (faslid2string x)))
                        (setq s (strlen x))
                  (binarywrite codeout* s)
                        (binarywriteblock codeout* (strbase x)
                                (strpack s))))
        (binarywrite codeout* % S is size in words
                     (setq s (WO (wplus2 currentoffset* 3))) )
        (binarywrite codeout* initoffset*)
        (binarywriteblock codeout* (WO codebase*) s)
        (binarywrite codeout*
                     (setq s
                      (wquotient
                       (wplus2 bittableoffset*
                        (wplus2 bittable-entries-per-word -1))
                       bittable-entries-per-word)))
        (binarywriteblock codeout* (WO bittablebase*) s)
      (deallocatefaslspaces)))
 
(de updatebittable (numberofentries firstentry)
  (let ((bto bittableoffset*) (btb bittablebase*))
   (when *writingfaslfile
    (putbittable btb bto firstentry)
    (setq bto (iadd1 bto))
    (for (from i 2 numberofentries 1)
         (do (progn (putbittable btb bto 0)
                    (setq bto (iadd1 bto)) )))
    (setq bittableoffset* bto  bittablebase* btb)
    (when (and (igreaterp bto (pa maxfasloffset*))
               (not (add-fasl-space)))       % try to enlarge
           (fatalerror
             "BPS exhausted during FaslOut; output file too large")))
) )
 
(de add-fasl-space ()   % try to enlarge fasl space
     (prog (new total btb temp)
       (setq btb bittablebase*)
       (when (not (eq lastbps (wo btb))) (return nil))
       (prin2t "+++ Enlarging fasl space +++")
  % cant do anything if faslspace is not on top of warray space
       (setq new (gtwarray 5000))
  % gtwarray will enlarge bps if necessary
       (setq btb bittablebase*)
       (setq total (wplus2 (WO (wdifference faslblockend* btb)) 5000))
 % move bittable
       (vector-copy (wo btb) new (wo (wdifference codebase* btb)))
       (setq bittablebase* (pa new))
       (setq temp (wplus2 (wplus2 bittablebase* 12)
                          (pa (wquotient total 9))))
 % move code
       (vector-copy (wo codebase*) (wo temp) (wplus2 maxfasloffset* 5))
       (setq codebase* temp)
       (setq maxfasloffset* (WO (wdifference faslblockend* codebase*)))
       (return
        (setq maxfasloffset* (Wplus2 maxfasloffset* -1))
)   ) )
 
%(de updatebittable (numberofentries firstentry)
 % (when *writingfaslfile
  %  (putbittable bittablebase* bittableoffset* firstentry)
   % (setq bittableoffset* (iadd1 bittableoffset*))
    %(for (from i 2 numberofentries 1)
     %    (do (progn (putbittable bittablebase* bittableoffset* 0)
      %              (setq bittableoffset* (iadd1 bittableoffset*)))))
  %  (when (igreaterp bittableoffset* maxfasloffset*)
   %  (fatalerror "BPS exhausted during FaslOut; output file too large"))))
 
(de allocatefaslspaces ()
  (prog (b)
    (setq b (gtwarray nil))
    % how much is left?
    (setq b (wdifference b (wquotient b 8)))
    (setq faslblockend* (PA (gtwarray 0)))
    % pointer to top of space
    (setq bittablebase* (PA (gtwarray b)))
    % take 7/8 of whatever's left
    (setq currentoffset* 0)
    (setq bittableoffset* 0)
%   (setq codebase*
%     (wplus2 bittablebase*               % split the space between
%          (PA (wquotient b               % bit table and code
%                      (WO bittable-entries-per-word)   ))))
    (setq codebase*
      (wplus2 (wplus2 bittablebase* 12)   % split the space between
           (PA (wquotient b 9 ))))        % bit table and code
    (setq maxfasloffset* (WO (wdifference faslblockend* codebase*)))
    (setq maxfasloffset* (Wplus2 maxfasloffset* -1))
    (setq orderedidlist* (cons nil nil))
    (setq nextidnumber* first-local-id-number)))
 
(de deallocatefaslspaces ()
  (delwarray (WO bittablebase*) (WO faslblockend*))
  )
 
(compiletime (remflag '(codefileheader codefiletrailer allocatefaslspaces
                  deallocatefaslspaces)
                  'internalfunction))
 
