%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC/MULS32.SL
% Description:  support for BIGNUM package with double INUM operations
%               lap source for HP RISC processor
% Author:       H. Melenk
% Created:      11 Sept  1989
% Modified:
% Mode:         Lisp
% Status:       Experimental
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Double length multiply and divide for the  processor.
% These are the open codes which embed the extended arithmetic 
% operations such that they are accessible from compiled LISP top level.
%
% Full 32 bit version : 28-June 1993 (H. Melenk)

(fluid '(*second-value* carry*))
 

(put 'wtimesdouble 'opencode
     % Called with two parameters of maximum bbase size
     % the return value are the low order bits of the product.
     % The high order bits are placed in fluid variable.
     '(% double length unsigned mutiply; 
       % EDX:EAX <- EAX * reg 
       (MUL (reg 2))
      % now we have 32 low bits in REG1, 32 high bits in REG4(=EDX)
       (*MOVE (reg EDX) ($FLUID *second-value*))
     ))

(put 'wxtimes2 'opencode % different version for $pxu/mbarith
     '((IMUL (reg 2))
      (*MOVE (reg 4)($FLUID *second-value*))
      ))

(put 'wquotientdouble 'opencode
      % called with a double length number in params 1 and 2
      % and a single length number in par 3.
      % Result is the single length quotient.
      % the remainder is placed in a fluid variable.
      '(  % adjusting the words first
        (*MOVE (reg 1) (reg edx)) % high word
        (*MOVE (reg 2) (reg eax)) % low word
        (DIV (reg 3))
        (*MOVE (reg EDX) ($FLUID *second-value*))
      ))

% add and set carry
%
%   (a + b) -> (carry*,result)

(put 'addAndSetCarry 'opencode
       '(
         (*MOVE (wconst 0)(reg 3))
         (ADD (reg 2)(reg 1))
           % move cf to carry*
         (ADDC (reg 3)(reg 3))
         (*MOVE (reg 3) ($FLUID carry*))
       ))

% add with carry
%
%   (a + b + carry*) -> (carry*,result)

(put 'addwithcarry 'opencode
       '(
           % move carry* to register CF
         (*MOVE ($fluid carry*)(reg 4))
         (*MOVE (wconst 0)(reg 3))
         (*MOVE (wconst 0)(reg 5))
         (SUB (reg 4)(reg 3))         
           % add with carry
         (ADDC (reg 2)(reg 1))
           % move cf to carry*
         (ADDC (reg 5)(reg 5))
         (*MOVE (reg 5) ($FLUID carry*))
       ))

% add and add carry
%
%   (a + b) -> result,  (*second-value* + carry) -> *second-value*

(put 'addAndAddCarry 'opencode
       '(
         (*move ($FLUID *second-value*) (reg 3))
         (*move 0 (reg 4))
           % add (reg 2) to (reg 1)
         (ADD (reg 2)(reg 1))
           % add carry to reg 3
         (ADDC (reg 4)(reg 3))
         (*move (reg 3) ($FLUID *second-value*))
       ))

% subtract with borrow
%
%  (a - (b + carry!*)) -> result, carry*=1 if borrow

(put 'subtractwithborrow 'opencode
       '(
           % move carry* to cf
         (*MOVE ($fluid carry*)(reg 4))
         (*MOVE (wconst 0)(reg 3))
         (*MOVE (wconst 0)(reg 5))
         (SUB (reg 4)(reg 3))    
           % subtract with borrow
         (SUBB (reg 2)(reg 1))
           % move new borrow to carry*
         (ADDC (reg 5)(reg 5))
         (*MOVE (reg 5) ($FLUID carry*))
       ))


%------------------- unsigned greaterp ---------------------

% code based on integer carry; 
 
(put 'ugreaterp* 'opencode 
   % returns 1 if arg1 > arg2 unsigned.
'( (*move (wconst 0) (reg 5))
   (sub (reg 1)(reg 2))        % compare, setting carry if r1>r2
   (addc (reg 5)(reg 5))       % move carry 
   (*move (reg 5)(reg 1))      % clear leading bits
))

(ds ugreaterp(a b)(eq 1 (ugreaterp* a b)))


