%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXK:SYS-SUPPORT.SL
% Description:  System specific kernel support functions.
% Author:       Brian Beach, Hewlett-Packard`CRC
% Created:      22-Feb-84
% Modified:     29-Jun-84 13:56:27 (RAM)
% Mode:         Lisp
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1989, Konrad-Zuse-Zentrum Berlin, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 28-Sep-89 (Winfried Neun)
%  Version for Mips
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(on fast-integers)

% mkcode etc look different for low tag scheme

(de mkcode (c)
       (prog (w)
          (setq w (gtheap 2))
          (putmem w (mkitem hcode-tag 0))
          (putmem (wplus2 w 1) c)
          (return (mkitem code-tag w))))

(lap '((*entry PlantUnbound expr 1)
       
	(*move (reg 1) (reg t2))
	(*WPLUS2 (reg t2) (fluid SYMFNC))       
	(lw (reg t1) UndefinedFunctionInstruction )
	(*move (reg t1) (displacement (reg t2) 0))
	(*EXIT 0)
    UndefinedFunctionInstruction
       (!.word "UNDEFINEDFUNCTION")))


(lap '( (*entry PlantCodePointer expr 2)
	(*move (reg 1) (reg t2))
	(*WPLUS2 (reg t2) (fluid SYMFNC))       
	(*move (reg 2) (displacement (reg t2) 0)) % and plant it
        (*EXIT 0)))


(lap '( (*entry PlantLambdaLink expr 1)
	(*move (reg 1) (reg t2))
	(*WPLUS2 (reg t2) (fluid SYMFNC))       
	(lw (reg t1) LambdaLinkInstruction )
	(*move (reg t1) (displacement (reg t2) 0))
	(*EXIT 0)
    LambdaLinkInstruction
       (!.word "COMPILEDCALLINGINTERPRETED")))


(lap '((*entry COMPILEDCALLINGINTERPRETED expr 0)
       (lw (reg t4) (plus2 SYMFNC (idloc COMPILEDCALLINGINTERPRETEDaux)))
       (jr (reg t4)))) % call aux without destroying t3

(lap '((*entry addressapply0 expr 1)
       (jr (reg 1))
       (nop)))
  
(de bittable (baseaddress bitoffset)
  (field (ilsh (byte baseaddress (ilsh bitoffset -2))
               (idifference (itimes2 (field bitoffset 30 2) 2) 6))
         30 2))

   % to be redefined in nonkernel

(lap '((*entry undefinedfunction expr 1)
       (lw (reg t4) (plus2 SYMFNC (times2 4 (idloc undefinedfunctionaux)))) 
       (jr  (reg t4)))) % call aux without destroying t3

(lap '((*entry undefinedfunctionaux expr 1)
       (*alloc 1)
       (*move (reg t3) (frame 1))
       (*move (quote "Undefined function called: ") (reg 1))
       (*call console-print-string)
       (*move (frame 1) (reg t1))
       (sll (reg t1) (reg t1) 2)
       (*wplus2 (reg t1) (fluid symnam))
       (*move (displacement (reg t1) 0) (reg 1))
       (and (reg 1) (reg 1) (reg infmask))
       (*call console-print-string)
       (*call console-newline)
       (*move 0 (reg 1))
       (*call Exit-with-status)
       (*exit 1)))

(de kernel-fatal-error (string)
  (console-print-string (kernelstring2string "FATAL ERROR: "))
  (console-print-string string)
  (console-newline)
  (exit-with-status -1))

(off fast-integers)
% signalhandler must reside in kernel

(lap '((*entry signalhandler expr 3)
%
% No alloc, trap handler will never come back!
%
%        (*move (indexed (reg 4) 272) (reg 5))
%        (*move 8#777774 (reg t1))
%        (*wnot (reg t1) (reg t1))
%        (*wand (reg 5) (reg t1))  % clear fp status mask
%        (*wor (reg 5) 8#3000)
%        (*move (reg 5) (indexed (reg 4) 272))

         (*movex (reg 1) (fluid Mips-signumber))
         (*movex (reg 2) (fluid Mips-sigcode))
         (*movex (reg 4) (fluid Mips-sigcp))   % save signal context
         (*move (indexed (reg 4) 148) (reg 4))
%        (*wshift (reg 4) -2)
         (*movex (reg 4) (fluid Mips-sigpc))

%  Check whether LISP trap handling is appropriate now

         (*movex ($fluid *kernelmode) (reg 5))
         (*jumpnoteq (label goon) (reg 5) 'nil)
         (*jumpnoteq (label templab) (reg 1) (quote 2)) % interrupt
         (*movex ($fluid *bruch*) (reg 5))
         (*jumpeq    (label templab) (reg 5) 0)
%
% Now we have a 'weaponed' interrupt. This is used e.g. by gc to remain
% operable. Return Immediately and mark with a negative value in *bruch* !!

%
    goon
         (*move -17 (reg 5))
         (*movex (reg 5) ($fluid *bruch*))
         (*move 0 (reg 5))
         (*movex (reg 5) (fluid Mips-sigcp))
         (jr (reg returnaddr))
         (nop)

    templab
         (*movex ($fluid Mips-sigcp) (reg t1))
         (*movex ($fluid Mips-dumparray*) (reg t2))
%        (*wshift (reg t2) 2)
         (*move 77 (reg t4))
      lab
         (*move (indexed (reg t1) 0) (reg t5))
         (*move (reg t5) (indexed (reg t2) 0))
         (*wplus2 (reg t1) 4)
         (*wplus2 (reg t2) 4)
         (*wplus2 (reg t4) -1)
         (*jumpwgreaterp (label lab) (reg t4) 0)

         (*movex  ($fluid mips-sigpc) (reg t4))
%        (*wshift (reg t4) 2)
         (*move (indexed (reg t4) 0) (reg t4))
         (*move 2#10101111101 (reg t5))
         (*wshift (reg t4) -21)
         (*jumpeq (label nodump) (reg t4) (reg t5))
         (*move 50 (reg t4))
         (*movex ($fluid Mips-dumparray*) (reg t1))
%        (*wshift (reg t1) 2)
         (*move (indexed (reg t1) 116) (reg t1))
         (*wdifference (reg t1) 40)
     stack-lab                % copy a centerdump of stack
         (*move (indexed (reg t1) 0) (reg t5))
         (*move (reg t5) (indexed (reg t2) 0))
         (*wplus2 (reg t1) 4)
         (*wplus2 (reg t2) 4)
         (*wplus2 (reg t4) -1)
         (*jumpwgreaterp (label stack-lab) (reg t4) 0)

     nodump
         (*jcall signal-done)
))

