/*  mini loader for PSL objects              				*/
/*				H. Melenk, ZIB Berlin			*/
 
/*  this version assumes, that pointers and long integers are   	*/
/*  compatible, e.g. as for 68020, 80386				*/
 
/* SECURITY_GAP defines a portion of memory which is held free
   for malloc; this gap has to be large enough such that the following
   BRK will lie in the free address space for all subsequent starts
   independent from the environment and other influences */
 
#define SECURITY_GAP   5000
 
/* RESERVE_GAP defines a portion of malloc space large enough such that
   the sbrk will not be called by (indirect) calls to malloc		*/

#define RESERVE_GAP    5000

#include <stdio.h>
#include "psll.h"
#include <sys/types.h>

extern char* malloc();

/*   #define DEFAULT "reduce.chkp" */
#define DEFAULT "main.b"

/*  directory for the subsequent nonkernel modules			*/
/*  ==============================================                      */
#define NKERN_DIR  "/keller/cons/psl/dist/nonkernel/sun/lap"
 
/*  processor specific values						*/
/*  =========================						*/
 
/*  PSL - to UNIX linkage 						*/
/*  =====================						*/

FILE * unix_stdin, * unix_stdout, * unix_stderr, * unix_tty;
int unix_null, unix_eof;

FILE * UNIXSTDIN, * UNIXSTDOUT, * UNIXSTDERR, * UNIXTTY;
int UNIXNULL, UNIXEOF;
 
 
#define     RELOC_TABLE -1

/*  bits in status word							*/
/*  ===================							*/
#define     STATUS_RELOCATABLE 1
#define     STATUS_KERNEL      3

/* structure definitions						*/
/* ===================== 						*/
 
struct file_head {
    long magic; /* magic number for consistency control 		*/
    long date;  /* date of production 					*/
    long type;  /* bits of file description				*/
    long sbrk;  /* old sbrk address 					*/
} fh;

struct segment_CB {
    long name;	/* name of the segment					*/
    long base;  /* ID number of the base value 				*/
    long type;	/* type of the information				*/
    long size;	/* size of segment in bytes				*/
    long low;	/* actual base address					*/
    long high;  /* actual high water mark				*/
};
 
struct segment_CB segment_list[SEGMENT_MAX];
long   segment_old_low[SEGMENT_MAX], 
       segment_old_high[SEGMENT_MAX],
       segment_diff[SEGMENT_MAX];
 
 
struct block_head {
    long name;	/* name of the corresponding segment			*/
    long rel;	/* load point (relative to segment low)			*/
    long len;	/* actual length of block				*/
    long rtype;	/* relocation type for this block			*/
} bh;
 
 
 /*  some global variables						*/
 /*  =====================						*/
 
long total_memory;
char * image = NULL;
char * nk_dir = NULL;
long  stack_size =0,
      heap_size = 0,
      bps_size =0,
      verbose = 0;

char * memory_base, * memory_next, * memory_high;

int segment_nbr;
int fd;         /* the file descriptor                                  */

long char2int(p)  /* convert ascii number into a long			*/
	char p[];
{	register long l; register int i; register char c;
	l=0;
	for(i=0; (c=p[i])!='\000';i++)
	  l = l*10 + (long)(c-'0');
	return(l);
}

para (argc,argv) 
   /* cracking of parameters */
	int argc; char* argv[];
{
	register int i; register char c; char * p;
	for(i=1; i<argc; i++)
	{c=argv[i][0];
	 if(c=='-')
		{c=argv[i][1]; p=&argv[i][2];
		 if(c=='i') image=p; 
		 else if(c=='d') nk_dir = p;
		 else if(c=='s') stack_size = char2int(p);
		 else if(c=='h') heap_size = char2int(p); 
		 else if(c=='b') bps_size = char2int(p);
		 else if(c=='v') verbose = char2int(p);
		}
	}
}
	
initial_memory()
{       long u,d; char*c; char*cr;
        if (fh.type & STATUS_KERNEL !=0)
	{    /* initial memory layout for a kernel */
		cr = malloc(RESERVE_GAP);
		c = malloc(SECURITY_GAP);
		memory_base = (char *) sbrk(total_memory);
		u = (long) memory_base;
		fh.sbrk = u;
		if(verbose>0) printf("memory base: %lx\n",memory_base);
		if (u<0) {printf ("not enough memory available\n");
		  exit(-1);};
		if ((u >> INFBITLENGTH)!=0 ||
			((u+total_memory)>> INFBITLENGTH)!=0)
			{printf ("memory addresses not comaptible with ");
			 printf ("PSL item format \n"); exit(-1);};
		free(c);
		free(cr);
	}
	else
	{   /* try to restore a previous memory layout */
		cr = malloc(RESERVE_GAP);
		memory_base = (char *) sbrk(0);
		u = (long) memory_base;
		if (u> fh.sbrk)
			if (fh.type & STATUS_RELOCATABLE==0)
			{ printf("PSLLOAD: free memory starts at %lx ",u);
			  printf(" but image requires start at %lx ",fh.sbrk);
			  printf(" and is NOT RELOCATABLE\n");
			  exit(-1); }
			else
			{ printf("PSLLOAD: relocatable savesystems not yet ");
			  printf("implemented\n"); exit(-1);};
		if((d=fh.sbrk-u)>0) memory_base = (char *) sbrk(d);
		memory_base = (char *) sbrk(total_memory); 
		free(cr);
                u = (long) memory_base; 
                if (u!=fh.sbrk)
			{printf ("memory allocation error: SBRK  ");
			 printf ("did not move with the expected amount\n");
			 exit (-1);};
                if(verbose>0) printf("memory base: %lx\n",memory_base);
                if (u<0) {printf ("not enough memory available\n");
		 exit(-1);};
	};
	memory_next = memory_base;
	memory_high = (char*)(u + total_memory); 
}


enlarge_memory(l)
  /* this routine is called when the memory is enlarged at runtime	*/
	long l;
{	long u; char*c;
	if ((((long)memory_high +l)>>INFBITLENGTH) !=0) 
                {printf("*** memory cannot be enlarged (2**%d <= %lx)\n", 
                        INFBITLENGTH,(long)memory_high +l); return(-1);};
	c = (char*)sbrk(l);
	if (c != memory_high)
	{printf("*** memory cannot be enlarged(discontinuous or exhausted)\n");
	 printf("*** old brk: %lx new: %lx\n",memory_high,c);
		return (-1); }
	u = (long)c;
	memory_high=(char*)(u + l);
	return(l);
}
	
long* p_malloc(n,k)
        long n,k;
{       long * u;
	u = (long*) memory_next;
	memory_next =(char*)((long) memory_next+n);
	if(verbose>0) printf("request (segment %ld) %lx   -> %lx\n",k,n,u);
        return(u);
}

main(argc,argv)
     int argc; 
     char * argv[];
     {long u; long * pu; char* a;
      int i,err;
	char * p;
      
      para(argc,argv);
      if (verbose>=1)
       printf("PSL loader parameters: %s %s %ld %ld %ld \n",
                       image, nk_dir,stack_size,heap_size,bps_size);
      if (image==NULL) image=DEFAULT; 
      fd = open(image,0);
      if (fd <0)
      { printf("image file %s cannot be opened for input\n",image);
        exit(-1);
       };
       
      read(fd,&fh,sizeof(fh));  /* magic */ 
      if (fh.magic!=MAGIC)
      { printf("file %s is not a legal PSL load file\n",argv[1]);
	printf("or PSLL was compiled with wrong header file\n");
        printf(" magic from header: %lx magic from load file: %lx\n",
		(long)MAGIC,fh.magic);
        exit(-1);};
 
      next_header(); 
      
      segment_nbr = bh.len / sizeof(segment_list[0]); /* number of segments*/
      if(verbose>=1) printf("%d segments\n",segment_nbr);
      read(fd,segment_list,bh.len);    
 
	/* update length information if possible*/
      if (fh.type & STATUS_KERNEL !=0) update_lengths();

      total_memory = 0;
      for (i=0; i<segment_nbr; i++)
		total_memory = total_memory+segment_list[i].size;
      if(verbose>=1) printf("total memory request: %ld(=%lxx) \n",total_memory);
      initial_memory();
                         
      /* now we have read in the segment table 		                */
      /* assign real memory to the segments,          			*/
      /* read data and do relocation					*/
       
      for (i=0; i<segment_nbr ; i++)
         if (i != SEGMENT_HEAP && i != SEGMENT_BPS)
		memory_for_segment(i);
	/*BPS and HEAP are assigned last for later relocatability	*/
	memory_for_segment(SEGMENT_BPS);
	memory_for_segment(SEGMENT_HEAP);
 
      /* read data blocks and do relocation         */
      for (i=1;i>0; i=load_block());
      for(i=0;i<segment_nbr; i++)
	{ 
	  /* put the start address into the cell of the base pointer    */
	  pu = (long*) (segment_list[SEGMENT_SYMVAL].low 
		        + segment_list[i].base * WORD_SIZE);
	  *pu = segment_list[i].low;
	  if(verbose>0) printf("segment %d starts at %lx (pointer in %lx)\n",
              i,segment_list[i].low,pu);
	};
     
	/* link to external variables					*/
      insert_fvars(segment_list[SEGMENT_SYMVAL].low);
      copy_bytes(frfntab,(char*)segment_list[SEGMENT_SYMFOREIGN].low,
		   sizeof(frfntab));
 
      if(nk_dir !=NULL) chdir(expand_file_name(nk_dir));

      /* execute program now */
      pu = (long*)segment_list[SEGMENT_START].low;
      if (verbose>0) printf("start address: %lx\n",*pu);
      fflush(stdout);
      execute(*pu,argc,argv);
      exit(0);
}     
       
memory_for_segment(i)
	int i;
{       char * a;
         /* preserve old memory addresses for relocation phase          */
         segment_old_low[i] = segment_list[i].low;
         segment_old_high[i]= segment_list[i].high;
 
         /* assign actual memory address                                */
         if (i==SEGMENT_CB) a = (char*) segment_list;
  /*     else if (i==SEGMENT_SYMFOREIGN) a = (char*) frfntab;  */
 
         else   a = (char*)p_malloc(segment_list[i].size,i);
 
         segment_diff[i] = (long)a - segment_list[i].low;
         segment_list[i].low = (long)a;
         segment_list[i].high = segment_list[i].high + segment_diff[i];
      }
          
int execute(start_addr,argc,argv)
	int (*start_addr)();
	int argc;
	char * argv[];
	{ int i;
	  i= (*start_addr)(argc,argv);  /* entry with original parameters */
          return(i);
	}
 
copy_bytes(from,to,lth)
	char from[]; char to[]; int lth;
{	int i;
	for (i=0; i<lth; i++) to[i]=from[i];
}

update_lengths()
	/* adjust lengths of segments from parameters */
        {
#ifdef SEGMENT_STACK
	 if (stack_size> segment_list[SEGMENT_STACK].high
                                -segment_list[SEGMENT_STACK].low+1)
                segment_list[SEGMENT_STACK].size = stack_size;
#endif
         if (bps_size> segment_list[SEGMENT_BPS].high
                                -segment_list[SEGMENT_BPS].low+1)
                segment_list[SEGMENT_BPS].size = bps_size;   
         if (heap_size> segment_list[SEGMENT_HEAP].high
                                -segment_list[SEGMENT_HEAP].low+1)
                segment_list[SEGMENT_HEAP].size = heap_size;  
        }

next_header()
{
	int i;
        i= read(fd,&bh,sizeof(bh));
	if (i<sizeof(bh)) read_error(i,sizeof(bh));
	if(verbose>0) printf("header:name %lx rel %lx len %lx rtype %lx\n",
		bh.name,bh.rel,bh.len,bh.rtype);
}  

load_block()
{
   long base_adr;
   int a_nr; int j;
   next_header();
   a_nr = bh.name;
   base_adr = segment_list[a_nr].low + bh.rel;
   if (bh.len ==0) return(0);                 /*ready now		*/
   if (bh.len > (j=read(fd, (char*)base_adr ,bh.len))) read_error(j,bh.len);
   reloc_block(base_adr, bh.len);
   return(1);
}
 
 reloc_block(b,l)
 	long b,l;
{
	if(bh.rtype==0) ;                      /* nothing to be done	*/
	else if(bh.rtype==-1) reloc_AA_items(b,l);
	else if(bh.rtype==-2) reloc_AA_jumps(b,l);
	else if(bh.rtype==-3) reloc_AA_words(b,l);
	else if(bh.rtype== 1) reloc_AA_table(b,l,1);
	else if(bh.rtype== 2) reloc_AA_table(b,l,WORD_SIZE);
	else {
		printf("unknown relocation type %ld\n",bh.rtype);
		exit(-1);}
	return(0);
}
 
int V;

long reloc_AA(l)
	/* do the relocation for one absolute address 			*/
	long l;
{
	register int i;
	register long l_inf;

        l_inf = l & INF_MASK;
	if (V==1) printf("reloc: %lx %lx ",l,l_inf);
	for (i=0;i<segment_nbr;i++)
	{ 
		if (l_inf>= segment_old_low[i] && l_inf<= segment_old_high[i])
			{l_inf = l_inf + segment_diff[i];
			 i = 9999;};
	}; 
	if (V==1) printf(" -->  %lx %lx\n", (l & TAG_MASK) | l_inf , l_inf);
	return((l & TAG_MASK) | l_inf);
}
		
long relocfromload(l)
	/* interface function for call from LISP */
	long l;
	{ return (reloc_AA(l)); }
		
long segmenttable(l)
	/* l is start address of a segment (or 0); return table address */
	long l;
	{int i; long r;
	 r=0;
	 if(l == -1) r =(long)(& fh);
	 else if(l==0) r =(long)(& segment_list[0].name);
	 else for(i=0; i<segment_nbr; i++)
	   if (l==segment_list[i].low) r=(long)(& segment_list[i].name);
	 return (r);
	}
 
long segmentlength(l) 
        /* l is start address of a segment (or 0); return segment length */ 
        long l; 
        {int i; long r; 
	 r=0;
         if(l==0) r =segment_list[0].size;  
         else for(i=0; i<segment_nbr; i++) 
           if (l==segment_list[i].low) r=(long)(segment_list[i].size); 
         return (r); 
        } 
  
 
reloc_AA_items(b,l)
	/* relocate a table of LISP items */
	long b,l;
{
	long * a,q; int tag;
	l = b+l;
	if (verbose>2) printf("reloc itmes; from %lx to %lx\n",b,l);
	for (b=b;b<l;b=b+WORD_SIZE)
	{       V = 0;
		a=(long*) b; q= *a;
                tag = (((unsigned )q) >> INFBITLENGTH);
		if (tag !=POSINT_TAG && tag != NEGINT_TAG && tag != ID_TAG )
		{   
			if (verbose>2){ printf("at %lx ",b); V=1;};
			*a = reloc_AA(q);
		}; 
	};
}
 
reloc_AA_words(b,l)
	/* relocate a table of absolute addresses */
        long b,l;
{
        long * a,q;
        l = b+l;
        if (verbose>2) printf("reloc words; from %lx to %lx\n",b,l);
        for (b=b;b<l;b=b+WORD_SIZE)
        {       V = 0;
                a=(long*) b; q= *a;
                if (verbose>2){ printf("at %lx ",b); V=1;};
                *a = reloc_AA(q);
        };
}
 

reloc_AA_jumps(b,l)
	long b,l;
{
	long *a,*aa;
	l = b+l;
        if (verbose>2) printf("reloc jumps; from %lx to %lx\n",b,l);
	for (b=b;b<l;b=b+JUMP_SIZE)
	{
		V = 0;
		if (verbose>2){printf("at %lx ",b); V=1;};
		a = (long*) (b+JUMP_OFFSET);
		*a = reloc_AA(*a);
	};

}
 
reloc_AA_table(b,l,d)
	long b,l;
	int d;
{
	long * a;
	long q;
	char * table,x;
	int i,j,table_len;
	next_header();
	if (bh.name != RELOC_TABLE)
	{ printf("expected relocation table not found\n"); exit(-1);};
	table_len = bh.len;
	if(bh.len < l/(8*d)) printf("table short: bh.len=%lx for l=%lx * %ld\n",bh.len,l,d);
	table = malloc(table_len);
	read(fd,table,table_len);
	
	V = 0;

	/* one bit in table coresponds to one data unit of size d-bytes	*/
	for (i=0; i<table_len; i++)
	{
		if ((x=table[i]) != 0)
		{ 
			for(j=0; j<8; j++)
			 if (1==((x>>(7-j))&1))
			   { 
				q = ((b + i*8 + j)*d);
				if (verbose>1){printf("at %lx ",q);V=1;};
				a=(long*)q;
			     	*a = reloc_AA(*a);
			};
		};
	};
	
	free(table); /* return table					*/
}
	
	
read_error(j1,j2)
	int j1,j2;
{  printf("***** too few data read: %d < %d  \n",j1,j2);
   exit(-1);}
