%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXK:MAIN-START.SL
% Title:          First routine called on startup
% Author:         Brian Beach, Hewlett-Packard CRC
% Created:        16-Feb-84
% Modified:       May 16 1985
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 2-Dec-88 (Winfried Neun)
%  Changed according to new calling sequence etc. (see "ZIB" comments)
% 31-Aug-88 (Julian Padget)
%  Changed fluids to globals, since latter needed in CPSL.
% 28-Oct-87 (Harold Carr)
%  Made panic exit call exit-with-status, rather than redundantely doing
%  what panic exit already did then calling external_exit.
% 13-Apr-87 (Harold Carr & Leigh Stoller)
%  Added external fluid _infbitlength_ to be used in bpsheap.c.
% 26-Sep-86 (Leigh Stoller)
%  Changed exit to unix to be a call to exit library function. The Sun does
%  accept the exit status in (reg 1) like the Bobcat. Called it external_exit
%  so it does not interfere with the cmacro of the same name.
% 22-Sep-86 (Leigh Stoller)
%  Commented out calls 68020_advise which is not defined on the sun.
% 18-Jul-86 (Leigh Stoller)
%  Changed the allocation of argumentblock to be the number of maxregs since
%   it is indexed by base + reg number.
% 15-Feb-85 13:27:53 ()
%  Made it so that PSL allocates the stack it will use.
%  Removed TST.B for stack expansion.
%  Included declaration of STACKSIZE (should go in SYS-CONSTS eventually).
% 9-Jul-84 09:11:55 (RAM)
%  Added fluids (and initialization) unixstdin, unixstdout, unixstderr,
%  unixnull, unixeof, unixtty.
% 3-Jul-84 10:51:27 (RAM)
%  Changed name of initial faslin file to $pnkl/load-psl.b since
%  binaryopenread can now expand shell variable references.
% 29-Jun-84 12:33:38 (RAM)
%  Chagged exit_to_unix to exit-to-unix to match call in kernel-fatal-error.
%  Removed save and restore of A6.
% 31-May-84 08:15:32 (Brian Beach)
%  Changed SYSCLEARIO call to HP_CLEARIO.  SYSCLEARIO is now called by the 
%  loader as part of its startup.
% 16-may-88 (chris burdorf and T. Yamamoto)
%  modified the main entry point so that it could handle our different
%  stack pointer and the sun-4's register windows to get arg c and argv from
%  the operating system.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime    %% Move this hack somewhere else later.
 (defmacro exported-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (flag1 id 'exportedsymbol)
         (put id 'symbol (kurzid id))
         (put (kurzid id) 'symbol (kurzid id))
         (put id 'vartype 'global)
         ))
   ))

(compiletime    %% Move this hack somewhere else later.
 (defmacro internal-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (flag1 id 'internalsymbol)
         (put id 'symbol (kurzid id))
	 (put (kurzid id) 'symbol (kurzid id))
         (put id 'vartype 'global)
         ))
   ))

(compiletime    %% Move this hack somewhere else later.
 (defmacro system-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (put id 'vartype 'global)
         ))
   ))

(compiletime (de declare-aux-2 (name upperbound)
     (findidnumber name)             % generate an ID it doesn't exist.
     (put name 'symbol (kurzid name))% flag as a fluid variable.
     (put  (kurzid name) 'symbol (kurzid name))
     (put name 'type 'fluid)         % flag as a fluid variable.
      (flag1 name 'externalsymbol)
	(setq name (kurzid name))
     (flag1 name 'externalsymbol)     % flag as initial symbol value.
     (when *declarebeforeuse (datadeclareexported name))
%     (dataalignfullword)
     (setq upperbound (list 'plus2 upperbound 1))
     (datareservezeroblock name upperbound)
     (unless *declarebeforeuse (datadeclareexported name))
%     (codedeclareexternal name)
))

(compiletime 
   (de no_kurzid (id) (prog (exp)
		(setq exp (explode2 id))
		(when (and (eq (car exp) 'U) (eq (cadr exp) 'N)
			   (eq (caddr exp) 'I) (eq (cadddr exp) 'X))
			(setq exp (cons 'u (cddddr exp))))
                (when (and (eq (car exp) 'O) (eq (cadr exp) 'L)
                           (eq (caddr exp) 'D) )
                        (setq exp (cons 'o (cdddr exp))))
                (when (and (eq (car exp) 'E) (eq (cadr exp) 'X)
				(greaterp (length exp) 9)
                           (eq (caddr exp) 'T)(eq (caddDr exp) 'E) )
                        (setq exp (cons 'E (pnth exp 9))))
		(setq exp (subst '@ '_ exp))
		(when (greaterp (length exp) 8)			
			(rplacd (pnth exp 8) nil))
		(return (intern (compress exp))))))

(on fast-integers)

(compiletime (load sys-consts))

(declare-kernel-structure 
 (stack             #.(compiler-constant 'stacksize))
 (argumentblock     #.(compiler-constant 'maxargs))
 (tokenbuffer       #.(compiler-constant 'maxtokensize))
 (bndstk            #.(compiler-constant 'bndstksize))
 (catchstack        #.(times2 (compiler-constant 'catchstacksize) 4))
 (hashtable         #.(quotient (compiler-constant 'hash-table-size) 2))
 (onewordbuffer     1)
 (saveargc          1)
 (saveargv          1)
 (datebuffer        1)
 )

(exported-global
 '(bndstkptr   bndstkupperbound   gcarraylowerbound   gcarrayupperbound
   heaplast   oldheaplast  nextsymbol
   heaplowerbound  heaptrapbound   oldheaptrapbound   heapupperbound
   _infbitlength_   lastbps   bpslowerbound   mainstartinitialize
   nextbps   oldheapupperbound   oldheaplowerbound
   stacklowerbound   stackupperbound   saveda7
   unixstdin   unixstdout   unixstderr   unixnull   unixeof   unixtty
   ))

(internal-global  '(   symnam  symval  symget symfnc  symprp   ))

(system-global  '(symnam symval symprp  symfnc symget
                  %  stack
                  catchstackptr   catchstack
                  bndstklowerbound   bndstkupperbound
                  bndstkptr   bndstk
                  heap
                  nextsymbol    %% Had SYMBOL as its assembler symbol.  why????
                    % Add these so they are always availbale to compiled code.
                  ***MUST-BE-NIL***   ***MUST-BE-NIL-TOO***
                  *fastcar
        %******************************************************
        %       SPARC specific  T.Yamamoto, July 12,1988
        %
                 vivarea    % ZIB
                 sparcjumpbuf
        %******************************************************
   ))

%(compiletime (copyd 'saveZIBanyregGlobal 'ZIBanyregGlobal))
(compiletime (de ZIBanyregGlobal (reg source) (list '$GLOBAL source)))

(de init-pointers()

  (setq catchstackptr catchstack)
  (setq bndstklowerbound bndstk)
  (setq bndstkupperbound (loc (wgetv bndstk (wdifference bndstksize 1))))
  (setq bndstkptr bndstk)
  (setq heap heaplowerbound))

(de init-fluids ()
        % Until problem of labels not appearing is fixed...
  (setq nextsymbol nil)
  (setq gcarraylowerbound nil)
  (setq gcarrayupperbound nil)
  (setq heaplast nil)
  (setq oldheaplast nil)
  (setq heaptrapbound nil)
  (setq oldheaptrapbound nil)
  (setq heapupperbound nil)
  (setq _infbitlength_ nil)
  (setq lastbps nil)
  (setq bpslowerbound nil)
  (setq mainstartinitialize nil)
  (setq nextbps nil)
  (setq oldheapupperbound nil)
  (setq oldheaplowerbound nil)
  (setq stackupperbound nil)
  (setq unixstdin nil)
  (setq unixstdout nil)
  (setq unixstderr nil)
  (setq unixnull nil)
  (setq unixeof nil)
  (setq unixtty nil)
  (setq ***MUST-BE-NIL*** NIL)
  (setq ***MUST-BE-NIL-TOO*** NIL)
  (setq *fastcar nil))

(compiletime (on mainfound))

(lap '((*entry pslmain expr 0)
       (*alloc 1) 

    % We had to change this code here a bit for the SUN-4, because
    % the sun-4 uses register windows to pass arguments and not a stack
    % as on the vax and 68xxx.  

%****************************************************************************

       %  Do OS specific initializations (uses argc and argv)
        (*move infbitlength (fluid _infbitlength_))
       (*link os_startup_hook expr 2)
       
       (*call init-pointers)

       (*move 128 (reg r9))
       (*mkitem (reg r9) id-tag)                    % initialize NIL reg
       (*move (reg r13) ($fluid stacklowerbound))
       (*call pre-main)                                 % call PSL

panic-exit                      % need to do UNIX cleanup after
                                % a fatal error, so jump here
       (*move (wconst 0) (reg 1))
       (*linke 1 exit-with-status expr 1)
       (*exit 1)

       (*entry exit-with-status expr 1)
	 (*alloc 0)
         (*move (reg 1) (fluid onewordbuffer))
         (*link os_cleanup_hook expr 0)
         (*move (fluid onewordbuffer) (reg 1))
         (*link external_exit expr 1)
         (*exit 1)))

(de pre-main ()
  %%(unixcleario)
  (initialize-symbol-table)
  (initcode)
  (console-print-string "About to load LOAD-PSL")
  (console-newline)
  (faslin "load-psl.b")
  (loader-main))

(off fast-integers)
