%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                     %
%  File:         Memory-sizes.sl                                      %
%  Description:  Adjust Sizes of Heap, Bps, Stack and Bndstk          %
%                       Cray COS Version                              %
%  Author : W. Neun , ZIB Berlin                                      %
%                                                                     %
%  Date   : 10-Feb-86                                                 %
%                                                                     %
%  Status : Experimental                                              %
%                                                                     %
% (c) Copyright 1987, Konrad Zuse Zentrum Berlin, all rights reserved %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%  Revisions:
%             8-Sep-88 W. Neun , new scheme for allocation (pairs in
%                      upper heap. Removed call for compacti-extension
%                      and the 8M case (handled by new compacti now)
%            24-Nov-87 W. Neun , added code for the 'rich' case:
%                      memory sizes beyond 8 M words
%                      stack and bndstk must not lay 'across' 8m bound
%            24-Sep-87 W. Neun , added new feature for the memory
%                      exhausted case
%            18-Nov-86 W. Neun , move some bps pointers when *writing
%                      faslfile, added add-heap-space
%            18-Aug-86 H.Melenk ZIB added load command for gc extension
%                      in case of more than 400000 words heap
%            25-Apr-86 added bndstk to dynamic allocated memory
%            27 June 86 added try-other-bps-spaces
 
%  Most of auxiliary routines (e.g. vector-copy, heap-manager contact)
%  are found in kernel(support) and/or getb00 (hard CAL)
 
(compiletime (load vector-instructions if-system)) %for clear-extraregs (in gc)
 
(compiletime (flag '(move-warray move-warray-temp-fix move-bndstk
                     move-stack-aux possible-system-error
                     possible-system-error1 possible-system-error2
                     set-heap-size-without-shrink
                     move-heap move-stack) 'internalfunction))
 
(fluid '(bndstk bndstkupperbound bndstklowerbound bndstkptr
         heap heaplast heaplowerbound heaptrapbound heapupperbound
         bps lastbps nextbps bpsupperbound
         stack stacklowerbound stackupperbound stkupbnd
         stacksize bpssize heapsize bndstksize *symcnt-allocated
         codebase* *writingfaslfile bittablebase* faslblockend*
         *AutoEnlargeMemory)
)

(setq *AutoEnlargeMemory t)

% First of all the Move-Routines
 
(De Move-Bndstk (from to)      % caution if counting is present
  (progn 
    (when (and (wlessp from to) *symcnt-allocated)
           (move-symcnt (wplus to bndstksize 3)))
    (vector-copy from to (wplus2 bndstksize 2))
    (setf bndstklowerbound to)
    (setf bndstk           to)
    (setf bndstkupperbound (wplus2 (wplus2 to bndstksize) -1))
    (setf bndstkptr (wplus2 bndstkptr (wdifference to from)))
    (update-catchstk (wdifference to from) 3)
      % is in catch-throw
    (when (and (wgreaterp from to) *symcnt-allocated)
           (move-symcnt (wplus to bndstksize 3)))
)   )
 
(De Move-Stack (from to)
  (progn 
         (vector-copy from to stacksize)
         (setf stacklowerbound to)
         (setf stack           to)
         (setf stackupperbound (wplus2 to stacksize))
         (setf stkupbnd (wdifference stackupperbound 200))
         (Move-stack-aux (wdifference to from))
         (update-catchstk (wdifference to from) 2)
)   )
 
(lap '((*entry move-stack-aux expr 1)
       (*move (reg 1) (reg a7))
       (*wplus2 (reg st) (reg a7))
       (*end 0)
)   )
 
(lap '((*entry set-stack-aux expr 0)
       (*move (reg st) (reg 1))
       (*end 0)
)   )
 
(De Move-heap (from to diff)
  (progn 
         (vector-copy from to heapsize)
         (setf heaplowerbound to)
         (setf heap           to)
         (setf heapupperbound (wplus2 to heapsize))
         (setf heaplast (wplus2 heaplast diff))
         (setf heaptrapbound (wplus2 heaptrapbound diff))
)   )
 
(compiletime   % macros for updating the warray-pointers that reside
 (progn        % in registers (B-regs)
               % the property *warray-pointer is set in CRFORMS
               % for assorted ids
 
  (fluid '(collectedregs*))
 
  (de collectregs-alter (x)
     (when (and (get x 'registername) (flagp x '*warray-pointer))
        (setq collectedregs*
         (cons `(setq ,x (wplus2 ,x diff)) collectedregs*))))
 
  (dm idtoreg-conversion-alter ()
       (setq collectedregs* nil)
       (mapobl (function collectregs-alter))
       (cons 'progn  collectedregs* ))
)) %end compiletime
 
(De Move-warray (from to lng)
     (prog (diff diff4)
         (setq diff (wdifference to from))
         (idtoreg-conversion-alter)
         (move-warray-temp-fix lambindargs*)
         (vector-copy from to lng)
         (when *writingfaslfile   % some goodies for faslout
              (setq diff4 (wtimes2 diff 4))
              (setq faslblockend* (wplus2 faslblockend* diff4))
              (setq codebase*     (wplus2 codebase*     diff4))
              (setq bittablebase* (wplus2 bittablebase* diff4)))
)    )
 
% very dirty hack to remain compatible after lambindargs* has gone to
% B66-Register
                        % update true fluid (not the reg)
(compiletime (setq *syslisp nil))
(de move-warray-temp-fix (la) (setq lambindargs* la))
(compiletime (setq *syslisp t))
 
(de set-bndstk-size (newsize)
  (prog (diff)
       (setq diff (wdifference newsize bndstksize))
       (cond
        ((and (wlessp diff 200) (wgreaterp diff -200))
            (prog2
             (print "Difference between old and new value is less 200")
             (return bndstksize)
       ))   )
       (when (possible-system-error1 diff)
              (print "***** Set-Bndstk-size impossible") (return NIL))
       (cond ((wlessp (wplus2 bndstkupperbound diff) bndstkptr)
              (conterror 4711
                "Can't shrink below actual bndstksize"
                (set-bndstk-size newsize)
       )     ))
       (cond ((wgreaterp diff 0)
                (if (enlarge-memory-block diff) nil
                 (progn
                  (set-heap-size-without-shrink diff)
                  (move-stack stack (wdifference stack diff))
                  (move-bndstk bndstk (wdifference bndstk diff))))
 
                (setq bndstkupperbound (wplus2 bndstkupperbound diff))
                (when  *symcnt-allocated
                    (move-symcnt (wplus2 bndstkupperbound 3)))
                (return (setq bndstksize newsize))
              )
              (t (progn
                  (setq bndstksize newsize)
                  (setq bndstkupperbound (wplus2 bndstkupperbound diff))
                  (when  *symcnt-allocated
                    (move-symcnt (wplus2 bndstkupperbound 3)))
                  (shrink-memory-block diff)
                  (return bndstksize)
)  )   )      )  )
 
(de set-stack-size (newsize)
  (prog (diff)
       (setq diff (wdifference newsize stacksize))
       (cond
        ((and (wlessp diff 1000) (wgreaterp diff -1000))
            (prog2
             (print "Difference between old and new value is less 1000")
             (return stacksize)
       ))   )
       (when (possible-system-error2 diff)
              (print "***** Set-Stack-size impossible") (return NIL))
       (cond ((wlessp (wplus2 stackupperbound diff) (set-stack-aux))
            (prog2
             (print "Can't shrink below actual Stacksize")
             (return stacksize)
       )    ))
       (cond ((wgreaterp diff 0)
                (if (enlarge-memory-block diff)
                  (move-bndstk bndstklowerbound
                           (wplus2 bndstklowerbound diff))
                 (progn
                  (set-heap-size-without-shrink diff)
                  (move-stack stack (wdifference stack diff))))
 
                (setf stackupperbound (wplus2 stackupperbound diff))
                (setf stkupbnd (wplus2 stkupbnd diff))
                (return (setq stacksize newsize))
              )
              (t (progn
                   (setq stacksize newsize)
                   (setf stackupperbound (wplus2 stackupperbound diff))
                   (setf stkupbnd (wplus2 stkupbnd diff))
                   (move-bndstk bndstklowerbound
                           (wplus2 bndstklowerbound diff))
                   (shrink-memory-block diff)
                   (return stacksize)
)  )   )      )  )
 
(de set-heap-size (newsize)
  (prog (diff)
       (setq diff (wdifference newsize heapsize))
       (cond
        ((and (wlessp diff 10000) (wgreaterp diff -10000))
            (prog2
             (print "Difference to actual value is less 10000")
             (return heapsize)
       ))   )
       (while (possible-system-error diff)
              (print "***** Set-Heap-size impossible") (return NIL))
       (cond ((wgreaterp heaplast (wplus2 heaptrapbound diff))
              (!%reclaim)))
       (cond ((and (wgreaterp heaplast (wplus2 heaptrapbound diff))
                   (wgreaterp heapupperbound (wlshift 1 23)))
              (!%reclaim)))
       (cond ((wgreaterp heaplast (wplus2 heaptrapbound diff))
                (prog2
                  (print "Can't shrink heapsize below actual usage")
                  (return heapsize)
       )     )  )
       (cond ((wgreaterp diff 0)
                (if (enlarge-memory-block diff)
                  (progn
                   (move-bndstk bndstklowerbound
                           (wplus2 bndstklowerbound diff))
                   (move-stack stacklowerbound
                           (wplus2 stacklowerbound diff))
                   (clear-memory heapupperbound diff)
                   (!%relink diff heaptrapbound)
                   (vector-copy heaptrapbound
                             (wplus2 heaptrapbound diff)
                             (wdifference heapupperbound heaptrapbound))
                   (setq heapupperbound (wplus2 heapupperbound diff))
                   (setq heaptrapbound  (wplus2 heaptrapbound diff))
                   (return (setq heapsize newsize))
                  )
                   (error 17 " Heap cannot be enlarged ")
                   (return heapsize)
              ) )
              (t (progn
                   (setq heapsize newsize)
                   (!%relink diff heaptrapbound)
                   (vector-copy heaptrapbound
                             (wplus2 heaptrapbound diff)
                             (wdifference heapupperbound heaptrapbound))
                   (setq heapupperbound (wplus2 heapupperbound diff))
                   (setq heaptrapbound  (wplus2 heaptrapbound diff))
                   (move-stack stacklowerbound
                           (wplus2 stacklowerbound diff))
                   (move-bndstk bndstklowerbound
                           (wplus2 bndstklowerbound diff))
                   (shrink-memory-block diff)
                   (return heapsize)
)  )   )      )  )
 
(de set-heap-size-without-shrink (malus)
  %
  % reducing heapsize by diff cells WITHOUT returning the memory
  % to the system.
  %
  (prog (diff newsize)
       (setq diff (wdifference 0 malus))
       (while (possible-system-error diff)
              (print "***** Set-Heap-size impossible") (return NIL))
       (setq newsize (wdifference heapsize malus))
       (prin2l (list "Reducing heapsize by" malus "cells"))
       (terpri)
       (cond ((wgreaterp heaplast (wplus2 heaptrapbound diff))
              (!%reclaim)))
       (cond ((wgreaterp heaplast (wplus2 heaptrapbound diff))
                (prog2
                  (print "Can't shrink heapsize below actual usage")
                  (return (error 17 "Heap reduction impossible"))
       )     )  )
       (setq heapsize newsize)
       (!%relink diff heaptrapbound)
       (vector-copy heaptrapbound (wplus2 heaptrapbound diff)
                         (wdifference heapupperbound heaptrapbound))
       (setq heapupperbound (wplus2 heapupperbound diff))
       (setq heaptrapbound  (wplus2 heaptrapbound diff))
       (return heapsize)
)  )
 
(de set-bps-size (newsize)
  (prog (diff)
       (setq diff (wdifference newsize bpssize))
       (cond
        ((and (wlessp diff 5000) (wgreaterp diff -5000))
            (prog2
             (print "Difference between old and new value is less 5000 ")
             (return bpssize)
       ))   )
       (while (possible-system-error diff)
              (print "***** Set-Bps-size impossible") (return NIL))
       (cond ((wgreaterp nextbps (wplus2 lastbps diff))
              (prog2
               (print "Can't shrink bps below current utilization")
               (return bpssize)
       )     ))
       (cond ((wgreaterp diff 0)
                (if (enlarge-memory-block diff)
                  (progn
                     (move-bndstk bndstklowerbound
                             (wplus2 bndstklowerbound diff))
                     (move-stack stacklowerbound
                             (wplus2 stacklowerbound diff)))
                  (set-heap-size-without-shrink diff))
 
                (!%relink diff lastbps)
                (clear-extraregs)
                (move-heap heaplowerbound
                          (wplus2 heaplowerbound diff) diff)
                (move-warray lastbps (wplus2 lastbps diff)
                             (wdifference bpsupperbound lastbps))
                (setq lastbps (wplus2 lastbps diff))
                (setq bpsupperbound (wplus2 bpsupperbound diff))
                (return (setq bpssize newsize))
              )
              (t (progn
                   (!%relink diff lastbps)
                   (move-warray lastbps (wplus2 lastbps diff)
                             (wdifference bpsupperbound lastbps))
                   (setq bpsupperbound (wplus2 bpsupperbound diff))
                   (setq bpssize newsize)
                   (setq lastbps (wplus2 lastbps diff))
                   (move-heap heaplowerbound
                           (wplus2 heaplowerbound diff) diff)
                   (move-stack stacklowerbound
                           (wplus2 stacklowerbound diff))
                   (move-bndstk bndstklowerbound
                           (wplus2 bndstklowerbound diff))
                   (shrink-memory-block diff)
                   (return bpssize)
)  )   )      )  )
 
(de try-other-bps-spaces (numberofitems)
   (prog (sizetoenlarge oldvalue outputbase* val)
          (setq outputbase* 10)
          (when (not *AutoEnlargeMemory) (return nil))
          (setq sizetoenlarge (wplus2 numberofitems 4999))
          (setq sizetoenlarge (wdifference sizetoenlarge
                              (wremainder sizetoenlarge 5000)))
          (setq sizetoenlarge (wplus2 sizetoenlarge bpssize))
          (setq oldvalue bpssize)
          (terpri)
          (prin2l
           (list "*** Size of BPS will be automatically enlarged to"
                         sizetoenlarge "items"))
          (terpri)
          (while (not (setq val (set-bps-size sizetoenlarge)))
          % 8 Mega word case, returns NIL if problems arise
                 (setq sizetoenlarge (wplus2 sizetoenlarge 10000)))
          (return (not (eq val oldvalue)))
)  )
 
(de add-heap-space (numberofitems)
   (prog (sizetoenlarge oldvalue outputbase* val)
          (setq outputbase* 10)
          (when (not *AutoEnlargeMemory) (return nil))
          (setq sizetoenlarge (wplus2 numberofitems 9999))
          (setq sizetoenlarge (wdifference sizetoenlarge
                              (wremainder sizetoenlarge 10000)))
          (setq sizetoenlarge (wplus2 sizetoenlarge heapsize))
          (setq oldvalue heapsize)
% must use emergency I/O
          (terpri) (console-newline)
          (console-print-string
            "*** Size of HEAP will be automatically enlarged to")
             (putint sizetoenlarge) (console-print-string " items")
          (while (not (setq val (set-heap-size sizetoenlarge)))
          % 8 Mega word case, returns NIL if problems arise
                 (setq sizetoenlarge (wplus2 sizetoenlarge 10000)))
          (return (not (eq val oldvalue)))
)  )
 
(if_system Y-MP
(de possible-system-error(diff) NIL)
(if_system Cray-2
(de possible-system-error(diff) NIL)
(de possible-system-error(diff)
 %
 % look at memory boundaries and check whether stack or bndstk
 % will be moved across the 8 M boundary (a rare case).
 % It is impossible (in the moment) to split stack and bndstk
 % at that boundary. heap may be split.
 %
  (prog (!8Mega)
     (setq !8Mega (wlshift 1 23))
     (return
        (or (and (wlessp (wplus2 bndstklowerbound diff) !8Mega)
                 (wgeq (wplus2 bndstkupperbound diff) !8Mega))
            (and (wlessp (wplus2 stack diff) !8Mega)
                 (wgeq (wplus2 stackupperbound diff) !8Mega))))))
))
(if_system Y-MP
(de possible-system-error1(diff) NIL) 
(if_system Cray-2 
(de possible-system-error1(diff) NIL)
(de possible-system-error1(diff)
 %
 % look at memory boundaries and check whether stack or bndstk
 % will be moved across the 8 M boundary (a rare case).
 % It is impossible (in the moment) to split stack and bndstk
 % at that boundary. heap may be split.
 %
  (prog (!8Mega)
     (setq !8Mega (wlshift 1 23))
     (return
        (or (and (wlessp bndstklowerbound !8Mega)
                 (wgeq (wplus2 bndstkupperbound diff) !8Mega))
            (and (wlessp (wplus2 stack diff) !8Mega)
                 (wgeq (wplus2 stackupperbound diff) !8Mega))))))
))
 
(if_system Y-MP
(de possible-system-error2(diff) NIL) 
(if_system Cray-2 
(de possible-system-error2(diff) NIL)
(de possible-system-error2(diff)
 %
 % look at memory boundaries and check whether stack or bndstk
 % will be moved across the 8 M boundary (a rare case).
 % It is impossible (in the moment) to split stack and bndstk
 % at that boundary. heap may be split.
 %
  (prog (!8Mega)
     (setq !8Mega (wlshift 1 23))
     (return
        (or (and (wlessp (wplus2 bndstklowerbound diff) !8Mega)
                 (wgeq (wplus2 bndstkupperbound diff) !8Mega))
            (and (wlessp stack !8Mega)
                 (wgeq (wplus2 stackupperbound diff) !8Mega))))))
))
