%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:PRINTERS.SL
% Title:        Printing functions for various data types
% Author:       Eric Benson
% Created:      27 August 1981
% Modified:     29-Aug-84 10:07:49 (Brian Beach)
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:  PL:IO-DECLS.B PL:TOKEN-DECLS.B PL:SYS-CONSTS.B PL:SYS-MACROS.B
% Runtime:
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 27-Jan-95 (Herbert Melenk)
%  Introduced function >output-case< for supporting upper case output.
% 03-NOV-94 (Herbert Melenk)
%  Installed switching to lower case PSL **low-case.
% 19-Sep-87 10:00 (Winfried Neun, ZIB)
%  Cray specific version WITHOUT string halfwords
% split up checklinefit
% 19-Jul-84 10:00 (Brian Beach)
%  Added (STRINF ...) around uses of DIGITSTRING in STRBYT.
% 12-Jul-84 10:00 (Brian Beach)
%  Added compile-time load of sys-macros for warray declarations.
% 31-May-84 10:46:35 (Brian Beach)
%  Call on IDAPPLY2 --> IDAPPLY.
% 10-May-84 14:19:21 (Brian Beach)
%  <PSL.KERNEL>PRINTERS.RED.18,  6-Feb-84 10:30:27, Edit by KESSLER
%  As Per Hearn - Floating Point right offset is too large.  Changed from
%   30 to 14.
%
% 22-Mar-84 11:42:42 (Brian Beach)
%  Added compiletime load of io-decls.
% 01-Dec-83 14:57:36 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (load io-decls token-decls sys-consts sys-macros))

(compiletime (flag '(charneedsescape output-switch-case) 'internalfunction))
 
(on fast-integers)
 
(fluid '(outputbase* % current output base
       prinlength % length of structures to print
       prinlevel % level of recursion to print
       currentscantable*
       lispscantable*
       idescapechar*
       *lower    % print IDs with uppercase chars lowered
       **low-case % lower case PSL
       output-case*  % eq 'raise: print IDs with lowercase chars raised
       ))
 
 
         (setq outputbase* 10)
         (setq idescapechar* 33)    % (char !!)
         (setq currentscantable* lispscantable*)   % so tokentypeofchar works right
 
(compiletime
  (progn (ds uppercasep (ch) (and (wgeq ch (char !A)) (wleq ch (char !Z))))
         (ds lowercasep (ch) (and (wgeq ch (char !a)) (wleq ch (char !z))))
         (ds raisechar (ch) (wplus2 (wdifference ch (char !a)) (char !A)))
         (ds lowerchar (ch) (wplus2 (wdifference ch (char !A)) (char !a)))
       ))
 
(compiletime
  (flag '(checklinefit writenumber1 channelwritebitstring
          channelwritebitstraux
          checklinefit&channelwritestring checklinefit&channelwritechar
          checklinefit&channelwritestring-hardcase
          checklinefit&channelwritechar-hardcase)
        'internalfunction))
 

(de output-case(u)
  (let((c output-case*))
   (when (and u (not (eq u 'raise)))
         (typeerror u 'output-case "a supported mode"))
   (setq output-case* u)
   c))


%. Writes EOL first if given Len causes max line length to be exceeded
(de checklinefit (len chn fn itm)
  (when (and (wgreaterp (wplus2 (wgetv lineposition chn) len)
                        (wgetv maxline chn))
             (wgreaterp (wgetv maxline chn) 0))
    (channelwritechar chn (char eol)))
  (idapplye2 chn itm fn))
 
(de checklinefit&channelwritestring-hardcase (len chn itm)
  (channelwritechar chn (char eol))
  (channelwritestring  chn itm))
 
(de checklinefit&channelwritechar-hardcase (len chn itm)
  (channelwritechar chn (char eol))
  (channelwritechar  chn itm))
 
(de checklinefit&channelwritestring (len chn itm)
 (prog() 
  (if (and (wgreaterp (wplus2 (wgetv lineposition chn) len)
                        (wgetv maxline chn))
             (wgreaterp (wgetv maxline chn) 0))
    (return (checklinefit&channelwritestring-hardcase len chn itm)))
 (return (channelwritestring  chn itm))))
 
(de checklinefit&channelwritechar (len chn itm)
 (prog() 
  (if (and (wgreaterp (wplus2 (wgetv lineposition chn) len)
                        (wgetv maxline chn))
             (wgreaterp (wgetv maxline chn) 0))
    (return (checklinefit&channelwritechar-hardcase len chn itm)))
  (return (channelwritechar chn itm))))
 
(de channelwritestring (channel strng)
  %
  % Strng may be tagged or not, but it must have a length field accesible
  % by StrLen.
  %
  (prog (uplim)
        (setq uplim (strlen strng))
        (for (from i 0 uplim 1)
              (do (channelwritechar channel (strbyt strng i))))         ))
 
(de writestring (s)
  (channelwritestring out* s))
 
(fluid '(digitstring))
 
(setq digitstring "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ")
 
(compiletime (fluid '(writenumberbuffer)))
 
(setq writenumberbuffer (gtwarray 14))
(clear-wstring  writenumberbuffer 100 14)
 
(compiletime (ds i-powerof2p (x)
             (if (eq 1 (wpopcnt x)) (wdifference 63 (wldzero x)) nil)))
 
(de channelwritesysinteger (channel number radix)
  (let ((exponent (i-powerof2p radix)))
 
  (cond (exponent
         (channelwritebitstring channel number (wdifference radix 1)
                          exponent))
        ((wlessp number 0)
         (channelwritechar channel (char '!-))
     %   (writenumber1 channel (wminus (longquotient number radix))
     %             radix)
     %   % To catch largest NEG, BUT ON CRAY 1/X we cannot divide
     %     LARGEST NEG !!!
     %   (channelwritechar channel
     %                 (strbyt digitstring
     %                               (wminus (wremainder number radix)))))
         (writenumber1 channel (wminus number) radix))
        ((weq number 0)
         (channelwritechar channel (char !0)))
        (t
         (writenumber1 channel number radix)))))
 
(de writenumber1 (channel number radix)
 (prog (remaind quotien)
  (setq quotien (longdivide-quotient (longdivide number radix)))
  (setq remaind (longdivide-remainder))
  (return
    (if (weq number 0)
        channel
        (progn (writenumber1 channel quotien radix)
               (channelwritechar channel
                                       (strbyt digitstring remaind)
))) )   )      )
 
(de channelwritebitstring (channel number digitmask exponent)
  (if (weq number 0)
    (channelwritechar channel (char !0))
    (channelwritebitstraux channel number digitmask exponent)))
 
(de channelwritebitstraux (channel number digitmask exponent)
  (cond ((weq number 0) channel)
        (t % Channel means nothing here                                     % just trying to fool the compiler
 
           (progn (channelwritebitstraux channel
                   (wrshift number exponent) digitmask exponent)
                  (channelwritechar channel
                            (strbyt digitstring (wand number digitmask)))))))
 
(de writesysinteger (number radix)
  (channelwritesysinteger out* number radix))
 
(de channelwritefixnum (channel num)
  (channelwriteinteger channel (fixval (fixinf num))))
 
(de channelwriteinteger (channel num)
  (when (wneq outputbase* 10)
    (channelwritesysinteger channel outputbase* 10)
    (channelwritechar channel (char !#)))
  (channelwritesysinteger channel num outputbase*)
  )
 
(de channelwritesysfloat (channel floatptr)
  (prog (ch chindex)
        (writefloat writenumberbuffer floatptr)
        (channelwritestring channel writenumberbuffer)))
 
(de channelwritefloat (channel lispfloatptr)
  (channelwritesysfloat channel (floatbase (fltinf lispfloatptr))))
 
(de channelprintstring (channel strng)
  (prog (len ch)
    (channelwritechar channel (char !"))
    (setq len (strlen strng))
    (for (from i 0 len 1)
       (do (progn (setq ch (strbyt strng i))
                  (when (weq ch (char !"))
                         (channelwritechar channel (char !")))
                  (channelwritechar channel ch))))
    (channelwritechar channel (char !"))))
 
(de output-switch-case(ch)
  (if **low-case
       (if (lowercasep ch) (raisechar ch) ch)
       (if (uppercasep ch) (lowerchar ch) ch)))

(de channelwriteid (channel itm)
  (cond ((or (and **low-case (not (eq output-case* 'raise)))
             (and (not **low-case) (not *lower)))
         (channelwritestring channel (symnam (idinf itm))))
        (t
    (prog (ch len)
      (setq itm (strinf (symnam (idinf itm))))
      (setq len (strlen itm))
      (for (from i 0 len 1)
           (do (progn (setq ch (output-switch-case (strbyt itm i)))
                      (channelwritechar channel ch))))))))
 
(de channelwriteunbound (channel itm)
  (channelwritestring channel "#<Unbound:")
  (channelwriteid channel itm)
  (channelwritechar channel (char '>)))
 
(de charneedsescape(ch)
   (or (and (null **low-case) (lowercasep ch))
       (and **low-case (uppercasep ch))))

(de channelprintid (channel itm)
  (prog (len ch tokentype)
    (setq itm (strinf (symnam (idinf itm))))
    (setq len (strlen itm))
    (setq ch (strbyt itm 0))
    (when (or (wneq (tokentypeofchar ch) 10) (charneedsescape ch))
      (channelwritechar channel idescapechar*))
    (if (or (and **low-case (not (eq output-case* 'raise)))
            (and (not **low-case) (not *lower)))
      (progn (channelwritechar channel ch)
             (for (from i 1 len 1)
                  (do
                   (progn (setq ch (strbyt itm i))
                          (setq tokentype (tokentypeofchar ch))
                          (unless (or (wleq tokentype 10)
                                      (weq tokentype plussign)
                                      (weq tokentype minussign))
                            (channelwritechar channel idescapechar*))
                          (when (charneedsescape ch)
                            (channelwritechar channel idescapechar*))
                          (channelwritechar channel ch)))))
      (progn (channelwritechar channel (output-switch-case ch))
             (for (from i 1 len 1)
                  (do
                   (progn (setq ch (strbyt itm i))
                          (setq tokentype (tokentypeofchar ch))
                          (unless (or (wleq tokentype 10)
                                      (weq tokentype plussign)
                                      (weq tokentype minussign))
                            (channelwritechar channel idescapechar*))
                          (when (charneedsescape ch)
                            (channelwritechar channel idescapechar*))
                          (setq ch (output-switch-case ch))
                          (channelwritechar channel ch))))))))

(de channelprintunbound (channel itm)
  (channelwritestring channel "#<Unbound ")
  (channelprintid channel itm)
  (channelwritechar channel (char '>)))
 
(de channelwritecodepointer (channel cp)
  (prog (n)
        (setq cp (codeinf cp))
        (channelwritestring channel "#<Code ")
        (setq n (!%code-number-of-arguments cp))
        (when (and (wgeq n 0) (wleq n maxargs))
          (channelwritesysinteger channel n 10)
          (channelwritechar channel (char blank)))
        (channelwritesysinteger channel cp compressedbinaryradix)
        (channelwritechar channel (char '>))))
 
(de channelwriteunknownitem (channel itm)
  (channelwritestring channel "#<Unknown ")
  (channelwritesysinteger channel itm compressedbinaryradix)
  (channelwritechar channel (char >)))
 
(de channelwriteblankoreol (channel)
  (if (and (wgeq (wplus2 (wgetv lineposition channel) 1)
                 (wgetv maxline channel))
           (wgreaterp (wgetv maxline channel) 0))
    (channelwritechar channel (char eol))
    (channelwritechar channel (char ! ))))
 
(de channelwritepair (channel itm level)
  (if (and (intp prinlevel) (wgeq level prinlevel))
    (channelwritechar channel (char '!#))
    (prog (n)
          (setq level (wplus2 level 1))
          (checklinefit&channelwritechar 1 channel (char !( ))
          (if (or (not (intp prinlength)) (wleq 1 prinlength))
            (progn (recursivechannelprin2 channel (car itm) level)
                   (setq n 2)
                   (setq itm (cdr itm))
                   (while (and (pairp itm)
                           (or (not (intp prinlength)) (wleq n prinlength)))
                     (channelwriteblankoreol channel)
                     (recursivechannelprin2 channel (car itm) level)
                     (setq n (wplus2 n 1))
                     (setq itm (cdr itm)))
                   (cond ((pairp itm)
                          (checklinefit&channelwritestring 3 channel
                           " ..."))
                         (itm
                          (checklinefit&channelwritestring 3 channel
                           " . ")
                          (recursivechannelprin2 channel itm level))))
            (checklinefit&channelwritestring 3 channel  "..."))
          (checklinefit&channelwritechar 1 channel (char !))))))
 
(de channelprintpair (channel itm level)
  (if (and (intp prinlevel) (wgeq level prinlevel))
    (channelwritechar channel (char '!#))
    (prog (n)
          (setq level (wplus2 level 1))
          (checklinefit&channelwritechar 1 channel (char !())
          (if (or (not (intp prinlength)) (wleq 1 prinlength))
            (progn (recursivechannelprin1 channel (car itm) level)
                   (setq n 2)
                   (setq itm (cdr itm))
                   (while (and (pairp itm)
                           (or (not (intp prinlength)) (wleq n prinlength)))
                     (channelwriteblankoreol channel)
                     (recursivechannelprin1 channel (car itm) level)
                     (setq n (wplus2 n 1))
                     (setq itm (cdr itm)))
                   (cond ((pairp itm)
                          (checklinefit&channelwritestring 3 channel
                           " ..."))
                         (itm
                          (checklinefit&channelwritestring 3 channel
                           " . ")
                          (recursivechannelprin1 channel itm level))))
            (checklinefit&channelwritestring 3 channel "..."))
          (checklinefit&channelwritechar 1 channel (char !))))))
 
(de channelwritevector (channel vec level)
  (if (and (intp prinlevel) (wgeq level prinlevel))
    (channelwritechar channel (char '!#))
    (prog (len i)
          (setq level (wplus2 level 1))
          (checklinefit&channelwritechar 1 channel (char '![))
          (setq len (veclen (vecinf vec)))
          (when (wlessp len 0)
            (return (checklinefit&channelwritechar
                                       1 channel (char '!]))))
          (setq i 0)
     loopbegin
          (if (or (not (intp prinlength)) (wlessp i prinlength))
            (progn (recursivechannelprin2 channel (vecitm (vecinf vec) i)
                    level)
                   (when (wleq (setq i (wplus2 i 1)) len)
                     (channelwriteblankoreol channel)
                     (go loopbegin)))
            (checklinefit&channelwritestring 3 channel "..."))
          (checklinefit&channelwritechar 1 channel (char  '!])))))
 
(de channelprintvector (channel vec level)
  (if (and (intp prinlevel) (wgeq level prinlevel))
    (channelwritechar channel (char '!#))
    (prog (len i)
          (setq level (wplus2 level 1))
          (checklinefit 1 channel 'channelwritechar (char '![))
          (setq len (veclen (vecinf vec)))
          (when (wlessp len 0)
            (return (checklinefit 1 channel 'channelwritechar (char '!]))))
          (setq i 0)
     loopbegin
          (if (or (not (intp prinlength)) (wlessp i prinlength))
            (progn (recursivechannelprin1 channel (vecitm (vecinf vec) i)
                    level)
                   (when (wleq (setq i (wplus2 i 1)) len)
                     (channelwriteblankoreol channel)
                     (go loopbegin)))
            (checklinefit 3 channel 'channelwritestring "..."))
          (checklinefit 1 channel 'channelwritechar (char '!])))))
 
(de channelwriteevector (channel evec level)
  (prog (handler)
        (cond ((and (intp prinlevel) (wgeq level prinlevel))
               (channelwritechar channel (char '!#)))
              ((and (getd 'object-get-handler-quietly)
                    (setq handler
                     (object-get-handler-quietly evec 'channelprin)))
               (apply handler (list evec channel level nil)))
              (t (channelwritestring channel "#<EVector ")
                 (channelwritesysinteger channel (evecinf evec)
                               compressedbinaryradix)
                 (channelwritechar channel (char '>)) nil))))
 
(de channelprintevector (channel evec level)
  (prog (handler)
        (cond ((and (intp prinlevel) (wgeq level prinlevel))
               (channelwritechar channel (char '!#)))
              ((and (getd 'object-get-handler-quietly)
                    (setq handler
                     (object-get-handler-quietly evec 'channelprin)))
               (apply handler (list evec channel level t)))
              (t (channelwritestring channel "#<EVector ")
                 (channelwritesysinteger channel (evecinf evec)
                               compressedbinaryradix)
                 (channelwritechar channel (char '>)) nil))))
 
(de channelwritewords (channel itm)
  (prog (len i)
        (channelwritestring channel "#<Words:")
        (setq len (wrdlen (wrdinf itm)))
        (when (wlessp len 0)
          (return (checklinefit 1 channel 'channelwritechar (char '>))))
        (setq i 0)
   loopbegin
        (if (or (not (intp prinlength)) (wlessp i prinlength))
          (progn (checklinefit 10 channel 'channelwriteinteger
                  (wrditm (wrdinf itm) i))
                 (when (wleq (setq i (wplus2 i 1)) len)
                   (channelwriteblankoreol channel)
                   (go loopbegin)))
          (checklinefit 3 channel 'channelwritestring "..."))
        (checklinefit 1 channel 'channelwritechar (char '>))))
 
(commentoutcode
(de channelwritehalfwords (channel itm)
  (prog (len i)
        (channelwritestring channel "#<Halfwords:")
        (setq len (halfwordlen (halfwordinf itm)))
        (when (wlessp len 0)
          (return (checklinefit 1 channel 'channelwritechar (char '>))))
        (setq i 0)
   loopbegin
        (if (or (not (intp prinlength)) (wlessp i prinlength))
          (progn (checklinefit 10 channel 'channelwriteinteger
                  (halfworditm (halfwordinf itm) i))
                 (when (wleq (setq i (wplus2 i 1)) len)
                   (channelwriteblankoreol channel)
                   (go loopbegin)))
          (checklinefit 3 channel 'channelwritestring "..."))
        (checklinefit 1 channel 'channelwritechar (char '>))))
) % commentoutcode
 
(de channelwritebytes (channel itm)
  (prog (len i)
        (channelwritestring channel "#<Bytes:")
        (setq len (strlen itm))
        (when (wlessp len 0)
          (return (checklinefit&channelwritechar 1 channel (char '>))))
        (setq i 0)
   loopbegin
        (if (or (not (intp prinlength)) (wlessp i prinlength))
          (progn (checklinefit 10 channel 'channelwriteinteger
                  (strbyt itm i))
                 (when (wleq (setq i (wplus2 i 1)) len)
                   (channelwriteblankoreol channel)
                   (go loopbegin)))
          (checklinefit&channelwritestring 3 channel "..."))
        (checklinefit&channelwritechar 1 channel (char '>))))
 
(de channelprin2 (channel itm)
  %. Display Itm on Channel
  (recursivechannelprin2 channel itm 0))
 
(de recursivechannelprin2 (channel itm level)
  (case (tag itm)
    ((posint-tag negint-tag)
     (checklinefit 10 channel 'channelwriteinteger itm))
    ((id-tag)
     (checklinefit (wplus2 (strlen (symnam (idinf itm))) 1)
               channel 'channelwriteid itm))
    ((unbound-tag)
     (checklinefit (wplus2 (strlen (symnam (idinf itm))) 12)
               channel 'channelwriteunbound itm))
    ((string-tag)
     (checklinefit&channelwritestring (wplus2 (strlen itm) 1) channel
               itm))
    ((code-tag)
     (checklinefit 14 channel 'channelwritecodepointer itm))
    ((fixnum-tag) (checklinefit 20 channel 'channelwritefixnum itm))
    ((floatnum-tag) (checklinefit 14 channel 'channelwritefloat itm))
    ((words-tag) (channelwritewords channel itm))
    ((halfwords-tag) (channelwriteunknownitem channel itm))
    ((bytes-tag) (channelwritebytes channel itm))
    ((pair-tag) (channelwritepair channel itm level))
    ((vector-tag) (channelwritevector channel itm level))
    ((evector-tag) (channelwriteevector channel itm level))
    (nil (checklinefit 20 channel 'channelwriteunknownitem itm)))
  itm)
 
(de prin2 (itm)
  %. ChannelPrin2 to current channel
  (recursivechannelprin2 out* itm 0))
 
(de channelprin1 (channel itm)
  %. Display Itm in READable form
  (recursivechannelprin1 channel itm 0))
 
(de recursivechannelprin1 (channel itm level)
  (case (tag itm)
    ((posint-tag negint-tag)
     (checklinefit 10 channel 'channelwriteinteger itm))
    ((id-tag) % leave room for possible escape chars
     (checklinefit (wplus2 (strlen (symnam (idinf itm))) 5)
               channel 'channelprintid itm))
    ((unbound-tag) % leave room for possible escape chars
     (checklinefit (wplus2 (strlen (symnam (idinf itm))) 16)
               channel 'channelprintunbound itm))
    ((string-tag)
     (checklinefit (wplus2 (strlen itm) 4) channel
               'channelprintstring itm))
    ((code-tag)
     (checklinefit 14 channel 'channelwritecodepointer itm))
    ((fixnum-tag) (checklinefit 20 channel 'channelwritefixnum itm))
    ((floatnum-tag) (checklinefit 20 channel 'channelwritefloat itm))
    ((words-tag) (channelwritewords channel itm))
    ((halfwords-tag) (channelwriteunknownitem channel itm))
    ((bytes-tag) (channelwritebytes channel itm))
    ((pair-tag) (channelprintpair channel itm level))
    ((vector-tag) (channelprintvector channel itm level))
    ((evector-tag) (channelprintevector channel itm level))
    (nil (checklinefit 20 channel 'channelwriteunknownitem itm)))
  itm)
 
(de prin1 (itm)
  %. ChannelPrin1 to current output
  (recursivechannelprin1 out* itm 0))
 
(off fast-integers)
 
