%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:RDS-WRS.SL
% Title:        Switch the current input or output channel
% Author:       Eric Benson
% Created:      28 August 1981
% Modified:     22-Mar-84 11:54:15 (Brian Beach)
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:  PL:IO-DECLS.B
% Runtime:
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 22-Mar-84 11:54:08 (Brian Beach)
%  Added compile time load of io-decls.
% 01-Dec-83 15:00:39 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (load io-decls))
 
(global '(specialrdsaction* % possibly apply to old and new channel
          specialwrsaction* % ditto
          in* % Current input channel
          out*)) % Current output channel
 
 
(fluid '(stdin* % Standard input - may be rebound
       stdout*))
 
% Standard output - may be rebound
(de rds (channel)
  % Switch input channels, return old
  (prog (oldin readfn)
        (when specialrdsaction*
          (apply specialrdsaction* (list in* channel)))
        (setq oldin in*)
        (when (null channel)
          (setq channel stdin*))
        (setq readfn (wgetv readfunction (intinf channel)))
        (when (or (weq readfn 'channelnotopen)
                  (weq readfn 'writeonlychannel))
          (return (channelerror channel
                   "Channel not open for input in RDS")))
        (setf in* channel)
        (return oldin)))
 
(de wrs (channel)
  % Switch output channels, return old
  (prog (oldout writefn)
        (when specialwrsaction*
          (apply specialwrsaction* (list out* channel)))
        (setq oldout out*)
        (when (null channel)
          (setq channel stdout*))
        (setq writefn (wgetv writefunction (intinf channel)))
        (when (or (weq writefn 'channelnotopen)
                  (weq writefn 'readonlychannel))
          (return (channelerror channel
                   "Channel not open for output in WRS")))
        (setf out* channel)
        (return oldout)))
 
 
