% 
%  oload.red - Load Unix object files into running PSL.
% 
% Author:	Russell D. Fish
% 		Computer Science Dept.
% 		University of Utah
% Date:	Mon May  3 1982
% Copyright (c) 1982 Russell D. Fish
% 

% Edit by Russ Fish, Sun Mar 14 1986
%  Load from .oload.bps and .oload.warray.  This avoids megabytes of zeros
%  on disk in the .oload.out file, which spanned the gap between the bps
%  allocation point and the warray allocation point.
% 24 Apr 1985 13:37:55 by Robert Virding
%  Declare CharactersPerWord as FLUID and set it to 4.
%  This is normally done when the compiler is loaded.
% Edit by Russ Fish, 23 Dec 1983
%  Fill in SymFunc cells as oloaded fns are put in the symbol table.
%  Also set current dir on .oload.out file path left in SymbolFileName!* .

% Symbols passed back from the oload script in .oload.syms.sl .
(fluid '(oload_text_origin oload_text_skip oload_text_size 
    oload_data_origin oload_data_skip oload_data_size oload_data_seg_size 
    oload_symtab_present))

% Symbol defined in the compiler.  Allow oloading without compiler loaded.
(fluid '(charactersperword))
(cond ((null charactersperword) (setq charactersperword 4)))

(ds wordcount (bytes)                   % Convert words to bytes.
    (quotient (plus bytes (difference charactersperword 1)) charactersperword))

(fluid '(*chattyoload))                 % Controls debug printing.
(setq *chattyoload nil)                 % Off by default now.

% oload - invokes the loader, making a number of .oload.* files on the
% connected directory, then pulls a successful result into the BPS area of
% the PSL.  Argument is a string containing arguments for the "oload"
% script, typically including a list of one or more .o and .a filename paths
% and flags to "ld", such as -l flags specifying libraries to search.  (See
% the oload script for a full description of its arguments.)
% 
% If the argument string is NIL, the oload script is skipped, and it is
% assumed that you are going to perform the same load over again.  This
% requires that the BPS and WARRRAY base addresses are the same as when the
% original oload was done.
%
% Return value is NIL if the oload failed, T if it worked.
%
(de oload (files) 
(prog (oload-nextbps oload-lastbps cmd chnl textsz 
	     datasegsz datasz bssbgn bssend) 

    (setq oload-nextbps (land (plus (gtbps 0) 3) -4)) 	% NextBPS, up to a word boundary.
    (setq oload-lastbps (land (gtwarray 0) -4)) 	% LastBPS, down to a word boundary.

    % Lots of debug print info, ignored unless variable set.
    (cond (*chattyoload (printf " NextBPS %p(%p), LastBPS %p(%p)%n" 
	oload-nextbps (gtbps 0) oload-lastbps (gtwarray 0)))) 

    % Load files into an a.out file.
    (cond (files 
    (progn 
    	(setq cmd (concat               % Avoid terpri in bldMsg if long Files string.
	    (bldmsg "oload %w %p %p " symbolfilename* oload-nextbps oload-lastbps) 
	    files)) 

	(cond (*chattyoload (print cmd))) 

	(system cmd)))) 
    

    % Get sizes and check for successful completion.
    (lapin ".oload.size.sl") 
    (cond ((eq oload_text_size -1) 
    (progn 
	(printf "oload: failed.%n") 
	(return nil) nil))) 
    

    % Skip to the end if only importing ForeignFunction knowledge of symbols
    % already linked into the running PSL via "-u" args.
    (cond ((and (eq oload_text_size 0) (eq oload_data_seg_size 0)) 
	(go finished))) 

    % Do some checking rather than blindly repeating a load.
    (setq bssend (plus oload_data_origin oload_data_seg_size)) 
    (cond ((and (not files) 
	(or (neq oload_text_origin oload-nextbps) (neq bssend oload-lastbps))) 
    (progn 
	(printf "oload: can't repeat load, file addresses don't match core.%n") 
	(printf "  Start (file %p, core %p), end (file %p, core %p).%n" 
	    oload_text_origin oload-nextbps bssend oload-lastbps) 
	(return nil) nil))) 
    

    % Position program text file past the text gap.
    (setq chnl (binaryopenread ".oload.bps")) 
    (binarypositionfile chnl oload_text_skip) 

    % Allocate storage space and read the text.
    (cond (*chattyoload (printf "bps %p, NextBPS %p, LastBPS %p%n" 
	(free-bps) (gtbps 0) (gtwarray 0)))) 
    (gtbps (plus 1 (setq textsz (wordcount oload_text_size)))) 
    (cond (*chattyoload 
    (progn 
	(printf "text %pw %pb %pa%n" textsz (times textsz 4) oload_text_origin) 
	(printf "bps %pw, NextBPS %p, LastBPS %p%n" 
	    (gtbps nil) (gtbps 0) (gtwarray 0)) nil))) 
    
    (binaryreadblock chnl oload_text_origin textsz) 
    (binaryclose chnl) 

    % Position past less than a page of gap at the start of initialized data.
    (setq chnl (binaryopenread ".oload.warray")) 
    (binarypositionfile chnl oload_data_skip) 

    % Allocate storage space for data and bss.
    (cond (*chattyoload (printf "bps %pw%n" (gtbps nil)))) 
    (setq datasegsz (wordcount oload_data_seg_size)) 	% Data + Bss
    (cond (*chattyoload (printf "data segment  %pw %pb%n" 
	datasegsz (times datasegsz 4)))) 
    (gtwarray (plus 1 datasegsz))       % Allocate for data and bss.
    (cond (*chattyoload (printf "bps %pw, NextBPS %p, LastBPS %p%n" 
	(gtbps nil) (gtbps 0) (gtwarray 0)))) 

    % Clear bss to zeros and read data.
    (setq bssbgn (plus oload_data_origin oload_data_size)) 
    (cond (*chattyoload (printf "bss start %p, end %p%n" bssbgn bssend))) 
    (clearblock bssbgn bssend) 
    (setq datasz (wordcount oload_data_size)) 	% Initialized data.
    (cond (*chattyoload (printf "data init  %pw %pb %pa%n" 
	datasz (times datasz 4) oload_data_origin))) 
    (binaryreadblock chnl oload_data_origin datasz) 
    (binaryclose chnl) 

 finished 
    % Load symbol definitions for the code and data just loaded.
    (lapin ".oload.syms.sl") 

    % Arrange that the symbol table for UnExec will come from the load file.
    (cond (oload_symtab_present 
	(setq symbolfilename* (concat (pwd) ".oload.symtab")))) 

    (return t)))                        % It worked.


% clearBlock - Clears allocated storage (for the BSS part of the oloaded pgm.)
(on syslisp)
(de clearblock (bgnaddr endaddr)
  (for (from i bgnaddr (wdifference endaddr 4) 4) (do (setf (getmem i) 0))))
(off syslisp)

% newForeignFn - Called by .oload.syms.sl to define function entries.
(on syslisp)                            % For mkCode.
(de newforeignfn (name loc) 
(prog nil 
    (put name 'foreignentry loc) 
    (makefcode name (mkcode loc)) 
    (flag1 name 'foreignfunction)))

(off syslisp)

% newForeignData - Called by .oload.syms.sl to define data entities.
(de newforeigndata (name loc sz) 
(prog nil 
    (put name 'foreigndata (cons loc sz))))

