%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:OPEN-CLOSE.SL 
% Title:        File primitives 
% Author:       Eric Benson 
% Created:      27 August 1981 
% Modified:     31-May-84 10:46:09 (Brian Beach) 
% Status:       Experimental 
% Mode:         Lisp 
% Package:      Kernel 
% Compiletime:  PL:IO-DECLS.B 
% Runtime:      
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(on fast-integers)

(compiletime (ds unixstring (x) (strbase (strinf x))))
(de unixstrlen (x) (hugo x (WRSHIFT x 61))) 
(DE hugo (x pad)   % count the length of a C string starting at PAD
   (PROG (l)    % this is together with hugo1 a  C2PSLstring
     (SETQ l pad)
     lll
     (WHEN (EQ 0 (BYTE x l)) (RETURN l))
     (SETQ l (WPLUS2 l 1))
     (GO lll)
   ))

(de socketopen(host socketnumber)

  (prog (filedespair filedes)
   (setq filedespair (systemopenfilesocket host socketnumber))
   (setq filedes (car filedespair))
   (setf (wgetv lineposition filedes) 0)
   (setf (wgetv maxline filedes) 80)
   (setf (wgetv unreadbuffer filedes) (char null))
   (setf (wgetv readfunction filedes)  'independentreadsocket)
   (setf (wgetv writefunction filedes) 'independentwritesocket)
   (setf (wgetv closefunction filedes) 'independentclosesocket)
   (setf (wgetv nextposition  filedes) 0)   % Will be post Incremented
   (setf (wgetv bufferlength  filedes) -1)
   (setf (wgetv maxbuffer filedes) (sysmaxbuffer (wgetv channeltable filedes)))
   (setf (igetv iobuffer  filedes) (mkstring (wgetv maxbuffer filedes) 32))

   (setq filedes (cdr filedespair)) 
   (setf (wgetv lineposition filedes) 0)
   (setf (wgetv maxline filedes) 80)
   (setf (wgetv unreadbuffer filedes) (char null))
   (setf (wgetv readfunction filedes)  'independentreadsocket)
   (setf (wgetv writefunction filedes) 'independentwritesocket)
   (setf (wgetv closefunction filedes) 'independentclosesocket)
   (setf (wgetv nextposition  filedes) 0)   % Will be post Incremented
   (setf (wgetv bufferlength  filedes) -1)
   (setf (wgetv maxbuffer filedes) (sysmaxbuffer (wgetv channeltable filedes)))
   (setf (igetv iobuffer  filedes) (mkstring (wgetv maxbuffer filedes) 32))
  (return  filedespair)))


(de systemopenfilesocket (host socketnumber)
 
  %  Mark a a pair of channels as open for a special purpose.
 
  (let ((channel (findfreechannel)) channel1)
    (setf (wgetv channelstatus channel) 'channelopenspecial)
    (setf (wgetv channeltable  channel) (syssocketopen host socketnumber))
    (setq channel1 (findfreechannel))
    (setf (wgetv channelstatus channel1) 'channelopenspecial)
    (setf (wgetv channeltable  channel1) (wgetv channeltable  channel))
    (cons channel channel1)
    ))
 
 
(de independentreadsocket (channel)

  % This function will read in a character from the buffer.  It will read
  % the record on buffer length overflow only.  Thus when an EOL character
  % is read, it is processed as any other character, except, if it is the last
  % one, in the record, it will do the read automatically.
  % Note, this will not read the next record until after the final character
  % has been processed.

  (testlegalchannel channel)

  % If the input buffer is empty, read another buffer full:

  (when (wgreaterp (wgetv nextposition channel) (wgetv bufferlength channel))
    (setf (wgetv bufferlength channel)
      (sysreadrecsocket (wgetv channeltable channel)
          (igetv iobuffer channel)))
    (setf (wgetv nextposition channel) 0))
 
  % Pull the next character out of the buffer.
 
  (let ((chr (strbyt (strinf (igetv iobuffer channel))
                     (wgetv nextposition channel))))
    (setf (wgetv nextposition channel) (+ (wgetv nextposition channel) 1))
    (when *echo
      (writechar chr))
    chr
    ))
 
(de independentwritesocket (channel chr)
 
  % Write a character into the buffer.  Actually dump the buffer when the
  % EOL character is found, or when the buffer is full.  This happens
  % immediately upon meeting this condition, not waiting for the
  % next character.  Note, that this places the EOL character into the
  % buffer for machine dependent treatment as CR/LF etc
 
  (testlegalchannel channel)
  (setf (wgetv nextposition channel) (+ (wgetv nextposition channel) 1))
  (setf (strbyt (strinf (igetv iobuffer channel))
                (wgetv nextposition channel))
    chr)
  (when (or (equal chr (char eol))
        (>= (add1 (wgetv nextposition channel))   % Indexed from 0
        (wgetv bufferlength channel)))        % Number of characters
    (socketflushbuffer channel)
  ) )

(de independentclosesocket (channel)

  (testlegalchannel channel)
  (unixclosesocket (wgetv channeltable channel))
  )

(de socketflushbuffer (channel)

  % Actually output any characters still hanging around in the buffer.

  (syswriterecsocket (wgetv channeltable channel)
           (igetv iobuffer channel)
           (wgetv nextposition channel))
  (setf (wgetv nextposition channel) -1)
  )


(de syssocketopen (host socketnumber)
  (prog (filed)
        (setq filed (unixsocketopen (if (stringp host)(unixstring host) host)
                                           socketnumber))
        (return (if (wleq filed unixnull)
                  (contopenerror host 'input)
                  filed))))

(de sysreadrecsocket (filedes stringbuffer)
    (prog nil
          (cond ((wneq (getsocket filedes (unixstring stringbuffer) 200)
                       unixnull)
                 % Check that not end of file.
                 (return (wdifference (unixstrlen stringbuffer) 1)))
                (t % Index of EOL char.
                   (progn % Return an EOF character on end of file.
                          (setf (strbyt (strinf stringbuffer) 0) $eof$)
                          (return 0))))))
 
(de syswriterecsocket (filedes stringtowrite stringlength)
  (writesocket filedes (unixstring stringtowrite) (wplus2 stringlength 1) ))
 
(de server ()
   (prog (files val)
       (setq files (socketopen 0 1188))
       (rds (car files)) 
     lupe
       (wrs nil)
       (setq val (eval (read)))
       (wrs (cdr files))
       (print (list 'prin2t (mkquote val)))
       (go lupe)))

(de client (str)
   (prog (files val) 
       (setq files (socketopen str 1188)) 
     lupe
       (rds nil) 
       (setq val (read))
       (rds (car files)) (wrs (cdr files)) 
       (print val)
       (wrs nil)
       (eval (read))
       (go lupe)
))

(de redserver ()
   (prog (files val)
       (setq files (socketopen 0 1188))
       (rds (car files)) 
     lupe 
       (wrs nil) 
       (setq val (aeval (read)))
       (wrs (cdr files)) 
       (print val) 
       (go lupe))) 
 
(de redclient (str)
   (prog (files val)
       (setq files (socketopen str 1188))
     lupe 
       (rds nil) 
       (setq val (xread nil))
       (rds (car files)) (wrs (cdr files))
       (print val)
       (wrs nil) 
       (setq val (read)) 
       (varpri val (mkquote val) 'only)
       (go lupe) 
)) 


(off fast-integers)
